"""
this module contains basic aeroval configurations for running on CI
"""

import copy
import logging
from pathlib import Path

from pyaerocom.io.pyaro.pyaro_config import PyaroConfig

logger = logging.getLogger(__name__)

# Constraints
DEFAULT_RESAMPLE_CONSTRAINTS = dict(
    yearly=dict(monthly=9),
    monthly=dict(
        daily=21,
        weekly=3,
    ),
    daily=dict(hourly=18),
)

DEFAULT_RESAMPLE_CONSTRAINTS_DAILY = dict(
    daily=dict(hourly=18),
)

HOMEDIR = Path.home()
MYPYAEROCOM_DIR = Path.home() / "MyPyaerocom"
TMP_DIR = Path.home() / "tmp"
JSON_DIR = Path.home() / "tmp" / "data"
COLDATA_DIR = Path.home() / "tmp" / "coldata"


# data directory for test data
TEST_DATA_DIR = MYPYAEROCOM_DIR / "testdata-minimal" / "obsdata" / "diurnal_test_data"

IO_AUX_FILE = MYPYAEROCOM_DIR / "testdata-minimal" / "config" / "gridded_io_aux.py"


def get_CFG(
    reportyear,
    year,
) -> dict:
    """create aeroval configuration dict to run the variable
    ratpm10pm25 (ratio pm10 vspm25)

    :returns: a dict of a model configuration usable for EvalSetup
    """
    MYPYAEROCOM_DIR.mkdir(exist_ok=True)
    TMP_DIR.mkdir(exist_ok=True)
    JSON_DIR.mkdir(exist_ok=True)
    COLDATA_DIR.mkdir(exist_ok=True)
    CFG = dict(
        json_basedir=JSON_DIR,
        coldata_basedir=COLDATA_DIR,
        io_aux_file=IO_AUX_FILE,
        # if True, existing colocated data files will be deleted and contours will be overwritten
        reanalyse_existing=True,
        only_json=False,
        add_model_maps=False,
        only_model_maps=False,
        modelmaps_opts=dict(maps_freq="monthly", maps_res_deg=5),
        clear_existing_json=False,
        # if True, the analysis will stop whenever an error occurs (else, errors that
        # occurred will be written into the logfiles)
        raise_exceptions=False,
        # Regional filter for analysis
        filter_name="ALL-wMOUNTAINS",
        # colocation frequency (no statistics in higher resolution can be computed)
        ts_type="hourly",
        map_zoom="Europe",
        freqs=[
            "monthly",
            "daily",
            "hourly",
        ],
        periods=[f"{year}"],
        main_freq="daily",
        zeros_to_nan=False,
        use_diurnal=True,
        min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
        colocate_time=False,
        obs_remove_outliers=False,
        model_remove_outliers=False,
        harmonise_units=True,
        regions_how="country",
        # annual_stats_constrained=True,
        proj_id="emepCI",
        exp_id=f"{reportyear}-reporting",
        exp_name=f"Evaluation of EMEP runs for {reportyear} EMEP reporting",
        exp_descr=(
            f"Evaluation of EMEP runs for {reportyear} EMEP reporting. The EMEP model, simulated for {year}, is compared against observations from EEA and EBAS."
        ),
        exp_pi="emep.mscw@met.no",
        public=True,
        # directory where colocated data files are supposed to be stored
        weighted_stats=True,
        var_order_menu=[
            # Gases
            "ratpm10pm25",
            "ratpm25pm10",
            "concNno",
            "concNno2",
            "concNtno3",
            "concNhno3",
            "concNtnh",
            "concNnh3",
            "concnh4",
            "concSso2",
            "concso4t",
            "concso4c",
            "vmro3",
            "vmro3max",
            "vmro3mda8",
            "vmrox",
            "vmrco",
            # PMs
            "concpm10",
            "concpm25",
            "concno3pm10",
            "concno3pm25",
            "concnh4pm25",
            "concso4pm25",
            "concCecpm10",
            "concCecpm25",
            "concCocpm10",  # SURF_ugC_PM_OMCOARSE missing in model-output
            "concCocpm25",
            "concsspm10",
            "concsspm25",
            # Depositions
            "wetrdn",
            "wetoxs",
            "wetoxn",
            "prmm",
        ],
    )

    CFG["model_cfg"] = {
        "EMEPCI": dict(
            model_id="EMEP.CI",
            model_ts_type_read="hourly",
        ),
    }

    """
    Filters
    """

    data_name = "CITestData"
    data_id = "harp"

    config = PyaroConfig(
        name=data_name,
        reader_id=data_id,
        filename_or_obj_or_url=TEST_DATA_DIR,
        filters={"variables": {"include": ["O3_density"]}},
        name_map={"O3_density": "vmro3"},
    )

    OBS_GROUNDBASED = {
        ################
        #    Pyaro
        ################
        "Pyaro-h": dict(
            obs_id=config.name,
            pyaro_config=config,
            web_interface_name=data_name,
            obs_name=data_name,
            obs_vars=["vmro3"],
            obs_vert_type="Surface",
            ts_type="hourly",
        ),
    }

    # Setup for supported satellite evaluations
    OBS_SAT = {}

    OBS_CFG = {**OBS_GROUNDBASED, **OBS_SAT}

    CFG["obs_cfg"] = OBS_CFG

    return copy.deepcopy(CFG)



from __future__ import annotations
from collections.abc import Iterable

import pandas as pd
import numpy as np
import pytest

from pyaerocom.io import ReadPyaro, PyaroConfig
from pyaerocom.io.pyaro.read_pyaro import PyaroToUngriddedData
from pyaerocom.io.pyaro.postprocess import matching_indices

from pyaerocom.ungridded_data_container import UngriddedDataContainer
from tests.conftest import lustre_unavail, __package_installed


def test_testfile(pyaro_test_data_file):
    assert pyaro_test_data_file.exists()


def test_readpyaro(pyaro_testdata):
    rp = pyaro_testdata

    assert isinstance(rp, ReadPyaro)


def test_variables(pyaro_testdata):
    rp = pyaro_testdata
    variables = ["NOx", "concso4", "od550aer", "NO", "PM10"]

    assert rp.PROVIDES_VARIABLES == variables
    assert rp.DEFAULT_VARS == variables


####################################################
#   Tests for the helper class PyaroToUngriddedData
####################################################


def test_pyarotoungriddeddata_reading(pyaro_testdata):
    from math import ceil

    obj = pyaro_testdata.converter
    data = obj.read()
    assert isinstance(data, UngriddedDataContainer)

    # Checks is data is empty
    assert not data.is_empty
    assert len(data.unique_station_names) == 2

    # Tests the found stations
    all_stations = data.to_station_data_all("concso4", ts_type_preferred="daily")

    assert all_stations["stats"][0]["ts_type"] in [
        "hourly",
        "3daily",
        "4daily",
        "2hourly",
        "2daily",
    ]
    assert all_stations["stats"][0]["country"] == "NO"

    # Tests the dates
    start = pd.to_datetime("01.01.2010", dayfirst=True)
    end = pd.to_datetime("31.12.2010", dayfirst=True)
    dates = pd.date_range(start, end, freq="D")
    assert len(all_stations["stats"][0].dtime) == ceil(len(dates) / 2)


def test_pyarotoungriddeddata_reading_kwargs(pyaro_testdata_kwargs):
    obj = pyaro_testdata_kwargs.converter
    data = obj.read()
    assert isinstance(data, UngriddedDataContainer)

    # Checks if stations have correct countries
    all_stations = data.to_station_data_all("concso4")
    countries = ["NO", "GB"]
    assert all_stations["stats"][1]["country"].strip() == countries[0]
    assert all_stations["stats"][0]["country"].strip() == countries[1]


def test_pyarotoungriddeddata_reading_extra_metadata(pyaro_testdata_kwargs):
    obj = pyaro_testdata_kwargs.converter
    data = obj.read()
    assert isinstance(data, UngriddedDataContainer)

    # Checks if stations have correct countries
    all_stations = data.to_station_data_all("concso4", add_meta_keys=["area_classification"])
    area_type = ["Rural", "Urban"]
    assert all_stations["stats"][1]["area_classification"].strip() == area_type[0]
    assert all_stations["stats"][0]["area_classification"].strip() == area_type[1]
    assert not isinstance(all_stations["stats"][0]["altitude"], Iterable)


def test_pyarotoungriddeddata_stations(pyaro_testdata):
    obj = pyaro_testdata.converter

    assert len(obj.reader.stations()) == 2


def test_pyarotoungriddeddata_variables(pyaro_testdata):
    obj = pyaro_testdata.converter

    assert obj.get_variables() == pyaro_testdata.PROVIDES_VARIABLES


def test_postprocessing(pyaro_test_data_file):
    config = PyaroConfig(
        reader_id="csv_timeseries",
        filename_or_obj_or_url=str(pyaro_test_data_file),
        name="test",
        name_map={
            "NO": "concno",
        },
        post_processing=[
            "concNno_from_concno",
        ],
        filters=dict(),
    )
    reader = ReadPyaro(config)
    assert set(reader.PROVIDES_VARIABLES).issuperset(["concno", "concNno"])
    data = reader.read(["concno", "concNno"])

    concno = data.all_datapoints_var("concno")
    concNno = data.all_datapoints_var("concNno")

    # Proportion of N in NO, ng -> ug conversion
    conversion_factor = 14.0067 / (14.0067 + 15.9994) * 1e-3

    assert np.allclose(concno * conversion_factor, concNno)


def test_matching_indices():
    x = [0, 1, 2, 3, 4]
    y = [1, 1.5, 2, 2.1, 4, 5]

    xind, yind = matching_indices(x, y)

    assert np.all(xind == [1, 2, 4])
    assert np.all(yind == [0, 2, 4])

    with pytest.raises(ValueError) as e:
        x = [6, 0, 1, 2, 3, 4]
        y = [1, 1.5, 2, 2.1, 4, 5]
        matching_indices(x, y)
    assert str(e.value) == "x is not monotonically increasing"


@lustre_unavail
@pytest.mark.skipif(
    not __package_installed("pyaro_readers"),
    reason="reader_id=eeareder requires pyaro-readers to be installed",
)
def test_vmrox():
    config = PyaroConfig.from_dict(
        {
            "name": "whatever",
            "reader_id": "eeareader",
            "filename_or_obj_or_url": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA-AQDS/download",
            "filters": {
                "countries": {"include": ["NO"]},
            },
            "name_map": {
                "O3": "conco3",
                "NO2": "concno2",
            },
            "post_processing": [
                "vmro3_from_conco3",
                "vmrno2_from_concno2",
                "vmrox_from_vmrno2_vmro3",
            ],
            "dataset": "unverified",
        }
    )
    reader = PyaroToUngriddedData(config)
    data = reader.read(vars_to_retrieve=["vmrox"])
    rev = data.get_data_revision("whatever")
    # WIP: waiting for eeareader to add revision, to be done by m06-2025
    # assert rev is not None
    rev is not None
    alldata = data.to_station_data_all()
    stats = alldata["stats"]
    assert len(stats) >= 4
    first = stats[0]
    assert first["units"] == {"vmrox": "nmol mol-1"}



from .tstype import TsType, sort_ts_types, get_lowest_resolution, get_highest_resolution
from .utils import (
    is_year,
    get_tot_number_of_seconds,
    to_datetime64,
    to_pandas_timestamp,
    infer_time_resolution,
    cftime_to_datetime64,
    datetime2str,
    seconds_in_periods,
)



from pathlib import Path

import pytest

from pyaerocom import UngriddedData
from pyaerocom.io import ReadAeronetSunV3
from pyaerocom.io.cachehandler_ungridded import CacheHandlerUngridded
from tests.conftest import lustre_avail
from pyaerocom.io.pyaro.pyaro_config import PyaroConfig
from pyaerocom.io.pyaro.read_pyaro import ReadPyaro


@pytest.fixture(scope="module")
def cache_handler():
    return CacheHandlerUngridded()


@lustre_avail
def test_cache_dir(cache_handler: CacheHandlerUngridded):
    cache_path = Path(cache_handler.cache_dir)
    comps = cache_path.parts
    assert comps[-2] == "_cache"
    assert comps[-3] == "MyPyaerocom"


def test_reload_custom(
    cache_handler: CacheHandlerUngridded, aeronetsunv3lev2_subset: UngriddedData, tmp_path: Path
):
    path = tmp_path / "test_manual_caching.pkl"
    cache_handler.write(aeronetsunv3lev2_subset, var_or_file_name=path.name, cache_dir=path.parent)
    assert path.exists()
    cache_handler.check_and_load(var_or_file_name=path.name, cache_dir=path.parent)
    assert cache_handler.loaded_data[path.name].shape == aeronetsunv3lev2_subset.shape


@pytest.mark.dependency
def test_reload(
    cache_handler: CacheHandlerUngridded,
    aeronetsunv3lev2_subset: UngriddedData,
    aeronet_sun_subset_reader: ReadAeronetSunV3,
):
    cache_handler.reader = aeronet_sun_subset_reader
    cache_handler.write(aeronetsunv3lev2_subset, var_or_file_name="od550aer")
    assert Path(cache_handler.file_path("od550aer")).exists()

    cache_handler.check_and_load(var_or_file_name="od550aer")
    subset = aeronetsunv3lev2_subset.extract_var("od550aer")
    assert "od550aer" in cache_handler.loaded_data

    reloaded = cache_handler.loaded_data["od550aer"]
    assert isinstance(reloaded, UngriddedData)
    assert reloaded.shape == subset.shape


def test_reload_pyaro(cache_handler: CacheHandlerUngridded, pyaro_test_data_file):
    config0 = PyaroConfig(
        reader_id="csv_timeseries",
        filename_or_obj_or_url=str(pyaro_test_data_file),
        name="test",
        name_map={
            "NO": "concno",
        },
        post_processing=[
            "concNno_from_concno",
        ],
        filters=dict(),
    )

    reader0 = ReadPyaro(config0)
    cache_handler.reader = reader0
    has_data = cache_handler.check_and_load("concno")
    assert not has_data, "Data was present"

    data = reader0.read("concno")
    _outpath = cache_handler.write(data, var_or_file_name="concno")

    has_data = cache_handler.check_and_load("concno")
    assert has_data, "Data should not be present"
    assert cache_handler.loaded_data.get("concno") is not None

    reader1 = ReadPyaro(config0)
    cache_handler.reader = reader1
    has_data = cache_handler.check_and_load("concno")
    assert has_data, "Data should be present"

    # Enable a filter which should invalidate the cache
    config2 = config0.model_copy()
    config2.filters["countries"] = {"include": "DE"}

    reader2 = ReadPyaro(config2)
    cache_handler.reader = reader2
    has_data = cache_handler.check_and_load("concno")
    assert not has_data, "Data should not be present"

    data = reader2.read("concno")
    _outpath = cache_handler.write(data, var_or_file_name="concno")
    has_data = cache_handler.check_and_load("concno")
    assert has_data, "Data should be present"



import logging
import os
import warnings

import numpy as np
import pandas as pd

from pyaerocom.units.datetime import get_tot_number_of_seconds
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.units import convert_unit

logger = logging.getLogger(__name__)


class ReadAasEtal(ReadUngriddedBase):
    """Read subset of GAW-TAD-EANET data related to Aas et al., 2019

    Paper URL: https://www.nature.com/articles/s41598-018-37304-0

    See Also
    ---------
    :class:`ReadUngriddedBase`
    """

    # name of files in GawTadSubsetAasEtAl
    _FILEMASK = "*.csv"  # fix

    #: version log of this class (for caching)
    __version__ = "0.09"

    COL_DELIM = ","

    #: Temporal resoloution
    TS_TYPE = "monthly"

    #: Name of dataset (OBS_ID)
    DATA_ID = "GAWTADsubsetAasEtAl"  #'GAWTADsubsetAasEtAl'

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: value corresponding to invalid measurement
    # NAN_VAL = -9999.
    NAN_VAL = -999.0
    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    # This contains the mapping between the requested variables and what it is called in the files.

    COLNAMES_VARS = {}
    COLNAMES_VARS["concentration_mgS/L"] = ["concso4pr"]
    COLNAMES_VARS["concentration_ugS/m3"] = ["concso4", "concso2"]
    COLNAMES_VARS["precip_amount_mm"] = ["pr"]
    COLNAMES_VARS["deposition_kgS/ha"] = ["wetso4"]

    #: Dictionary mapping filenames to available variables in the respective files.
    FILES_CONTAIN = {}
    FILES_CONTAIN["monthly_so2.csv"] = ["concso2"]
    FILES_CONTAIN["monthly_so4_aero.csv"] = ["concso4"]
    FILES_CONTAIN["monthly_so4_precip.csv"] = ["wetso4", "pr", "concso4pr"]

    #: Dictionary mapping variable name to hard coded filenames.
    VARS_TO_FILES = {}
    VARS_TO_FILES["concso2"] = ["monthly_so2.csv"]
    VARS_TO_FILES["concso4"] = ["monthly_so4_aero.csv"]
    VARS_TO_FILES["pr"] = ["monthly_so4_precip.csv"]
    VARS_TO_FILES["wetso4"] = ["monthly_so4_precip.csv"]
    VARS_TO_FILES["concso4pr"] = ["monthly_so4_precip.csv"]

    # (from to unit)
    UNITCONVERSION = {
        "concso2": ("ug S/m3", "ug m-3"),
        "concso4": ("ug S/m3", "ug m-3"),
        "wetso4": ("kg S/ha", "kg m-2"),  #  s-1
        "concso4pr": ("mg S/L", "g m-3"),
    }

    # =============================================================================
    #     VAR_UNITS_READ={}
    #     VAR_UNITS_READ['concso2'] = 'ug m-3'
    #     VAR_UNITS_READ['concso4'] = 'ug m-3'
    #     VAR_UNITS_READ['pr']       = 'mm'
    #     VAR_UNITS_READ['wetso4']   = 'kg m-2 s-1'
    #     VAR_UNITS_READ['concso4pr'] = 'g m-3' # Removed sulphur from VAR_UNITS_READ mgS/L
    # =============================================================================
    #: :obj: `list` of :obj: `str`
    #: List containing all the variables available in this data set.
    PROVIDES_VARIABLES = list(VARS_TO_FILES)

    #: int: Number of available variables in this data set.
    num_vars = len(PROVIDES_VARIABLES)

    @property
    def DEFAULT_VARS(self):
        """Default variables (wrapper for :attr:`PROVIDES_VARIABLES`)"""
        return self.PROVIDES_VARIABLES

    def _get_time_stamps(self, df):
        tconv = lambda yr, m: np.datetime64(f"{yr:04d}-{m:02d}-{1:02d}", "s")  # noqa: E731
        dates_alt = [tconv(yr, m) for yr, m in zip(df.year.values, df.month.values)]
        return np.asarray(dates_alt)

    def read_file(self, filename, vars_to_retrieve):  #  -> List[StationData]:
        """Read one GawTadSubsetAasEtAl file

        ToDo
        ----
        Could be more efficient...

        Note
        ----
        One file contains all stations for one variable. This is unctypical
        for ground based observations since the data files are typically by
        station.

        Parameters
        ----------
        filename : str
            absolute path to filename to read

        vars_to_retrieve : list or str, optional
            list containing variable IDs that are supposed to be read.

        Returns
        -------
        station_list : list
            list of `StationData` objects containing data
        """
        station_list = []
        df = pd.read_csv(filename, sep=",", low_memory=False)
        # Converting month and year.

        df["dtime"] = self._get_time_stamps(df)

        # array av numpy.datetime64
        df.rename(columns={"Sampler": "instrument_name"}, inplace=True)
        grouped = df.groupby(by="station_name")

        # Looping over every station:
        for name, station_group in grouped:
            station_group = station_group.drop_duplicates(
                subset="dtime", keep="first"
            )  # Drops duplicate rows

            tvals = station_group["dtime"]

            stat = StationData()
            # Meta data
            stat["station_name"] = name

            stat["filename"] = filename
            stat["data_id"] = self.data_id
            stat["ts_type"] = self.TS_TYPE

            var_names = self.COLNAMES_VARS
            # Looping over all keys in the grouped data frame.
            for key in station_group:
                # Enters if the the key is a variable
                if key in var_names:
                    _var = np.intersect1d(var_names[key], vars_to_retrieve)
                    if len(_var) == 0:
                        continue
                    elif len(_var) > 1:
                        raise OSError("Found multiple matches...")
                    var = _var[0]
                    if var in self.UNITCONVERSION:
                        # Convert units
                        from_unit, to_unit = self.UNITCONVERSION[var]
                        values = pd.to_numeric(station_group[key], errors="coerce").values
                        stat[var] = convert_unit(
                            values, from_unit=from_unit, to_unit=to_unit, var_name=var
                        )

                        if var == "wetso4":
                            # TODO: quite hard coded...
                            numsecs = get_tot_number_of_seconds(ts_type=self.TS_TYPE, dtime=tvals)
                            stat[var] = stat[var] / numsecs

                            to_unit = "kg m-2 s-1"
                    else:
                        to_unit = key.split("_")[-1]
                        vals = pd.to_numeric(station_group[key], errors="coerce").values
                        # This should only be true for
                        stat[var] = vals

                    stat["var_info"][var] = {"units": to_unit}
                else:
                    if key == "dtime":
                        stat[key] = station_group[key].values
                    else:
                        # Store the meta data.
                        stat[key] = station_group[key].values[0]

            # =============================================================================
            #             # Since no altitude is provided in the files. Geodesy is used to approximate
            #             # altitude based on latitude and longitude.
            #             try:
            #                 s["altitude"] = geodesy.get_topo_altitude(lat=s.latitude,
            #                                                           lon=s.longitude)
            #             except ValueError as e:
            #                 s['altitude'] = np.nan
            #                 from pyaerocom import const
            #                 logger.warning(f'Failed to access altitude for {name}')
            # =============================================================================
            # Added the created station to the station list.
            station_list.append(stat)
        return station_list

    def read(self, vars_to_retrieve=None):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or `str`:, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded

        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.

        Returns
        -------
        UngriddedData : :class:`UngriddedData`
            data object

        """

        files = self.get_file_list()
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        data_obj = UngriddedData()

        meta_key = 0.0
        idx = 0
        varindex = -1

        # assign metadata object
        metadata = data_obj.metadata  # dict
        meta_idx = data_obj.meta_idx  # dict

        for file in files:
            filename = os.path.basename(file)
            if filename not in self.FILES_CONTAIN:
                raise OSError(f"Invalid file name {filename}, this should not happen.")
            var_matches = [var for var in vars_to_retrieve if var in self.FILES_CONTAIN[filename]]
            if len(var_matches) == 0:
                continue
            stat_list = self.read_file(file, vars_to_retrieve=var_matches)
            for stat in stat_list:
                # self.counter += 1
                metadata[meta_key] = {}
                metadata[meta_key].update(stat.get_meta())
                metadata[meta_key].update(stat.get_station_coords())
                metadata[meta_key]["data_id"] = self.data_id
                metadata[meta_key]["ts_type"] = self.TS_TYPE
                # metadata[meta_key]['variables'] = stat["variables"]
                # Is instrumentname
                if "instrument_name" in stat and stat["instrument_name"] is not None:
                    instr = stat["instrument_name"]
                else:
                    instr = self.INSTRUMENT_NAME

                metadata[meta_key]["instrument_name"] = instr
                metadata[meta_key]["data_revision"] = self.data_revision

                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = {}

                num_times = len(stat["dtime"])
                num_vars = len(stat["var_info"])
                temp_vars = list(stat["var_info"])
                tconv = stat["dtime"].astype("datetime64[s]")
                times = np.float64(tconv)
                totnum = num_times * num_vars

                if (idx + totnum) >= data_obj._ROWNO:
                    # This results in a error because it doesn't want to multiply empty with nan
                    data_obj.add_chunk(totnum)

                metadata[meta_key]["var_info"] = {}
                for var_count, var in enumerate(temp_vars):
                    values = stat[var]
                    start = idx + var_count * num_times
                    stop = start + num_times

                    if var not in data_obj.var_idx:
                        varindex += 1
                        data_obj.var_idx[var] = varindex
                        var_idx = varindex
                    else:
                        var_idx = data_obj.var_idx[var]

                    metadata[meta_key]["var_info"][var] = stat["var_info"][var]

                    data_obj._data[start:stop, data_obj._LATINDEX] = stat["latitude"]
                    data_obj._data[start:stop, data_obj._LONINDEX] = stat["longitude"]
                    data_obj._data[start:stop, data_obj._ALTITUDEINDEX] = stat["altitude"]
                    data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key

                    # write data to data object
                    data_obj._data[start:stop, data_obj._TIMEINDEX] = times
                    data_obj._data[start:stop, data_obj._DATAINDEX] = values
                    data_obj._data[start:stop, data_obj._VARINDEX] = var_idx

                    meta_idx[meta_key][var] = np.arange(start, stop)

                meta_key += 1
                idx += totnum

        data_obj._data = data_obj._data[:idx]
        # sanity check
        data_obj._check_index()
        return data_obj


def _check_line_endings(filename):
    """File that checks all line endings in an ascii file (not in use)"""
    ll = None
    wrong_endings = {}
    with open(filename) as f:
        prev = None
        for i, line in enumerate(f.readlines()):
            spl = line.split(",")
            if not spl[-1] == "\n":
                wrong_endings[i + 1] = line
            if ll is None:
                ll = len(spl)
            elif not len(spl) == ll:
                print(i)
                print(prev)
                print(spl)
                raise Exception
            prev = spl
    return wrong_endings


class ReadSulphurAasEtAl(ReadAasEtal):
    """Old name of :class:`ReadAasEtal`."""

    def __init__(self, *args, **kwargs):
        super(ReadAasEtal, self).__init__(*args, **kwargs)
        warnings.warn(
            "You are using an old name for class ReadAasEtal", DeprecationWarning, stacklevel=2
        )



import abc
from fnmatch import fnmatch

from pyaerocom.aeroval.modelentry import ModelEntry
from pyaerocom.aeroval.obsentry import ObsEntry
from pyaerocom.exceptions import EntryNotAvailable


class BaseCollection(abc.ABC):
    def __init__(self):
        """
        Initialize an instance of BaseCollection.
        The instance maintains a dictionary of entries.
        """
        self._entries = {}

    def __iter__(self):
        """
        Iterates over each entry in the collection.

        Yields
        ------
        object
            The next entry in the collection.
        """
        yield from self._entries.values()

    @abc.abstractmethod
    def add_entry(self, key, value) -> None:
        """
        Abstract method to add an entry to the collection.

        Parameters
        ----------
        key: Hashable
            The key of the entry.
        value: object
            The value of the entry.
        """
        pass

    @abc.abstractmethod
    def remove_entry(self, key) -> None:
        """
        Abstract method to remove an entry from the collection.

        Parameters
        ----------
        key: Hashable
            The key of the entry to be removed.
        """
        pass

    @abc.abstractmethod
    def get_entry(self, key) -> object:
        """
        Abstract method to get an entry from the collection.

        Parameters
        ----------
        key: Hashable
            The key of the entry to retrieve.

        Returns
        -------
        object
            The entry associated with the provided key.
        """
        pass

    def keylist(self, name_or_pattern: str = None) -> list[str]:
        """Find model / obs names that match input search pattern(s)

        Parameters
        ----------
        name_or_pattern : str, optional
            Name or pattern specifying search string.

        Returns
        -------
        list
            list of keys in collection that match input requirements. If
            `name_or_pattern` is None, all keys will be returned.

        Raises
        ------
        KeyError
            if no matches can be found
        """
        # Special case where the model cfg is empty, used for obs-only
        if self._entries.keys() == []:
            return []

        if name_or_pattern is None:
            name_or_pattern = "*"

        matches = []
        for key in self._entries.keys():
            if fnmatch(key, name_or_pattern) and key not in matches:
                matches.append(key)

        if len(matches) == 0:
            raise KeyError(f"No matches could be found that match input {name_or_pattern}")
        return matches

    @property
    def web_interface_names(self) -> list:
        """
        List of web interface names for each obs entry

        Returns
        -------
        list
        """
        return self.keylist()

    def as_dict(self) -> dict:
        """
        Convert object to serializable dict

        Returns
        -------
        dict
            content of class

        """
        output = {}
        for key, val in self._entries.items():
            if hasattr(val, "json_repr"):
                val = val.json_repr()
            output[key] = val
        return output


class ObsCollection(BaseCollection):
    """
    Object that represents a collection of obs entries

    "Keys" are obs names, values are instances of :class:`ObsEntry`. Values can
    also be assigned as dict and will automatically be converted into
    instances of :class:`ObsEntry`.


    Note
    ----
    Entries must not necessarily be only observations but may also be models.
    Entries provided in this collection refer to the y-axis in the AeroVal
    heatmap display and must fulfill the protocol defined by :class:`ObsEntry`.

    """

    def add_entry(self, key: str, entry: dict | ObsEntry):
        if isinstance(entry, dict):
            entry = ObsEntry(**entry)
        self._entries[key] = entry
        self._entries[key].obs_name = self.get_web_interface_name(key)

    def remove_entry(self, key: str):
        if key in self._entries:
            del self._entries[key]

    def get_entry(self, key: str) -> ObsEntry:
        """
        Getter for obs entries

        Raises
        ------
        KeyError
            if input name is not in this collection
        """
        try:
            entry = self._entries[key]
            return entry
        except (KeyError, AttributeError):
            raise EntryNotAvailable(f"no such entry {key}")

    def get_all_vars(self) -> list[str]:
        """
        Get unique list of all obs variables from all entries

        Returns
        -------
        list
            list of variables specified in obs collection

        """
        vars = []
        for ocfg in self._entries.values():
            vars.extend(ocfg.get_all_vars())
        return sorted(list(set(vars)))

    def get_web_interface_name(self, key: str) -> str:
        """
        Get webinterface name for entry

        Note
        ----
        Normally this is the key of the obsentry in :attr:`obs_config`,
        however, it might be specified explicitly via key `web_interface_name`
        in the corresponding value.

        Parameters
        ----------
        key : str
            key of entry.

        Returns
        -------
        str
            corresponding name

        """
        entry = self._entries.get(key)
        return (
            entry.web_interface_name
            if entry is not None and entry.web_interface_name is not None
            else key
        )

    @property
    def web_interface_names(self) -> list:
        """
        List of web interface names for each obs entry

        Returns
        -------
        list
        """
        return [self.get_web_interface_name(key) for key in self.keylist()]

    @property
    def all_vert_types(self):
        """List of unique vertical types specified in this collection"""
        return list({x.obs_vert_type for x in self._entries.values()})


class ModelCollection(BaseCollection):
    """
    Object that represents a collection of model entries

    "Keys" are model names, values are instances of :class:`ModelEntry`. Values
    can also be assigned as dict and will automatically be converted into
    instances of :class:`ModelEntry`.

    Note
    ----
    Entries must not necessarily be only models but may also be observations.
    Entries provided in this collection refer to the x-axis in the AeroVal
    heatmap display and must fulfill the protocol defined by
    :class:`ModelEntry`.
    """

    def add_entry(self, key: str, entry: dict | ModelEntry):
        if isinstance(entry, dict):
            entry = ModelEntry(**entry)
        entry.model_name = key
        self._entries[key] = entry

    def remove_entry(self, key: str):
        if key in self._entries:
            del self._entries[key]

    def get_entry(self, key: str) -> ModelEntry:
        """
        Get model entry configuration
        Parameters
        ----------
        model_name : str
            name of model

        Returns
        -------
        dict
            Dictionary that specifies the model setup ready for the analysis
        """
        if key in self._entries:
            return self._entries[key]
        else:
            raise EntryNotAvailable(f"no such entry {key}")

    def keylist(self, name_or_pattern: str = None) -> list[str]:
        """Find model / obs names that match input search pattern(s)

        Parameters
        ----------
        name_or_pattern : str, optional
            Name or pattern specifying search string.

        Returns
        -------
        list
            list of keys in collection that match input requirements. If
            `name_or_pattern` is None, all keys will be returned.

        Raises
        ------
        KeyError
            if no matches can be found
        """
        # Special case where the model cfg is empty, used for obs-only
        if list(self._entries.keys()) == []:
            return []

        else:
            return super().keylist(name_or_pattern)



from __future__ import annotations

from pathlib import Path

import pytest

from pyaerocom.io.pyaro.pyaro_config import PyaroConfig


def get_test_config() -> PyaroConfig:
    config = PyaroConfig(
        name="test",
        reader_id="test",
        filename_or_obj_or_url="test",
        filters={},
        name_map={},
    )
    return config


def get_existing_config() -> PyaroConfig:
    config = PyaroConfig(
        name="aeronetsun_test",
        reader_id="test",
        filename_or_obj_or_url="test",
        filters={},
        name_map={},
    )
    return config


def test_default_path_exist():
    assert PyaroConfig._DEFAULT_CATALOG.exists()


def test_kwargs(pyaro_testconfig_kwargs, pyaro_kwargs):
    kwargs = pyaro_testconfig_kwargs.model_extra["columns"]
    assert kwargs == pyaro_kwargs


def test_save(tmp_path):
    config = get_test_config()
    config.save(path=Path(tmp_path))


def test_list_configs(tmp_path):
    config = get_test_config()
    config.save(path=Path(tmp_path))

    private_configs = list(PyaroConfig.load_catalog(tmp_path / "catalog.yaml").keys())
    default_configs = list(PyaroConfig.load_catalog().keys())

    configs = default_configs + private_configs

    assert configs == PyaroConfig.list_configs(tmp_path / "catalog.yaml")


def test_load(tmp_path):
    config = get_test_config()
    config.save(path=Path(tmp_path))

    new_config = PyaroConfig.load("test", filepath=tmp_path / "catalog.yaml")

    assert config == new_config


#########################################################
#   Tests for expected exceptions
#########################################################
def test_save_path_error(tmp_path):
    config = get_test_config()
    with pytest.raises(ValueError, match="must be a directory"):
        config.save(path=Path(tmp_path) / "catalog.yaml")


def test_save_duplicate_name_error(tmp_path):
    config = get_test_config()
    config.save(path=Path(tmp_path))
    with pytest.raises(ValueError, match="already exists in catalog"):
        config.save(path=Path(tmp_path))


def test_save_name_in_default_error(tmp_path):
    config = get_existing_config()
    with pytest.raises(ValueError, match="already exists in default"):
        config.save(path=Path(tmp_path))


def test_load_filepath_error(tmp_path):
    config = get_test_config()
    with pytest.raises(ValueError, match="is a directory not a file"):
        config.load("test", tmp_path)


def test_load_no_catalog_error(tmp_path):
    config = get_test_config()
    with pytest.raises(ValueError, match="No catalog with this name exists"):
        config.load("test", tmp_path / "catalog.yaml")


def test_load_no_config_error(tmp_path):
    config = get_test_config()
    config.save(path=Path(tmp_path))
    with pytest.raises(ValueError, match="was not found in any catalogs"):
        config.load("test2", tmp_path / "catalog.yaml")



import fnmatch
import logging
import os

from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom.variable import Variable
from pyaerocom.variable_helpers import parse_aliases_ini, parse_variables_ini

from pyaerocom.units import Unit

logger = logging.getLogger(__name__)


class VarCollection:
    """Variable access class based on variables.ini file"""

    def __init__(self, var_ini):
        self._all_vars = None
        self._var_ini = None

        self.var_ini = var_ini

        self._vars_added = {}

        self._cfg_parser = parse_variables_ini(var_ini)
        self._alias_parser = parse_aliases_ini()
        self._idx = -1

    @property
    def all_vars(self):
        """List of all variables

        Note
        ----
        Does not include variable names that may be inferred via
        alias families as defined in section [alias_families] in
        aliases.ini.
        """
        if self._all_vars is None:
            self._all_vars = list(self._cfg_parser) + list(self._vars_added)
        return self._all_vars

    @property
    def var_ini(self):
        """Config file specifying variable information"""
        return self._var_ini

    @var_ini.setter
    def var_ini(self, var_ini):
        if not isinstance(var_ini, str):
            raise ValueError("Invalid input for var_ini, need str")
        elif not os.path.exists(var_ini):
            raise FileNotFoundError(f"File {var_ini} does not exist")
        self._var_ini = var_ini

    def add_var(self, var):
        """Add a new variable to this collection

        Minimum requirement for new variables are attributes var_name and
        units.

        Parameters
        ----------
        var : Variable
            new variable definition

        Raises
        ------
        VariableDefinitionError
            if a variable is already defined under that name

        Returns
        -------
        None
        """
        if not isinstance(var.var_name, str):
            raise ValueError("Attr. var_name needs to be assigned to input variable")
        if var.var_name in self.all_vars:
            raise VariableDefinitionError(f"variable with name {var.var_name} is already defined")
        if not isinstance(var, Variable):
            raise ValueError("Can only add instances of Variable class...")
        if not isinstance(var.units, str):
            var.units = str(Unit(var.units))
        self._all_vars.append(var.var_name)
        self._vars_added[var.var_name] = var

    def delete_variable(self, var_name: str) -> None:
        """
        Remove input variable from this collection

        Parameters
        ----------
        var_name : str
            name of variable

        Raises
        ------
        VariableDefinitionError
            if variable does not exist or if it exists more than once.

        Returns
        -------
        None

        """
        all_vars = self.all_vars
        matches = [i for i, x in enumerate(all_vars) if x == var_name]
        if len(matches) == 0:
            raise VariableDefinitionError(f"No such variable {var_name} in VarCollection")
        elif len(matches) > 1:
            raise VariableDefinitionError(
                f"FATAL: found multiple matches for variable {var_name} in "
                f"VarCollection. Please check variables.ini"
            )
        all_vars.pop(matches[0])
        self._all_vars == all_vars
        if var_name in self._vars_added:
            del self._vars_added[var_name]

    def get_var(self, var_name):
        """
        Get variable based on variable name

        Parameters
        ----------
        var_name : str
            name of variable

        Raises
        ------
        VariableDefinitionError
            if no variable under input var_name is registered.

        Returns
        -------
        Variable
            Variable instance

        """
        if var_name in self._vars_added:
            return self._vars_added[var_name]
        var = Variable(var_name, cfg=self._cfg_parser)
        if var.var_name_aerocom not in self:
            raise VariableDefinitionError(
                f"Error (VarCollection): input variable {var_name} is not supported"
            )
        return var

    def find(self, search_pattern):
        """Find all variables that match input search pattern

        Note
        ----
        Searches for matches in variable names (:attr:`Variable.var_name`) and
        standard name (:attr:`Variable.standard_name`).

        Parameters
        ----------
        search_pattern : str
            variable search pattern

        Returns
        -------
        list
            AeroCom variable names that match the search pattern
        """
        matches = []
        for var in self:
            if fnmatch.fnmatch(var.var_name, search_pattern):
                matches.append(var.var_name)
            elif isinstance(var.standard_name, str) and fnmatch.fnmatch(
                var.standard_name, search_pattern
            ):
                matches.append(var.var_name)
        return matches

    def __dir__(self):
        """Activates auto tab-completion for all variables"""
        return self.all_vars

    def __iter__(self):
        return self

    def __next__(self):
        self._idx += 1
        if self._idx == len(self.all_vars):
            self._idx = -1
            raise StopIteration
        var_name = self.all_vars[self._idx]
        return self[var_name]

    def __contains__(self, var_name):
        if var_name in self.all_vars:
            return True
        return False

    def __len__(self):
        return len(self.all_vars)

    def __getitem__(self, var_name):
        return self.get_var(var_name)

    def __repr__(self):
        return f"VarCollection ({len(self)} entries)"

    def __str__(self):
        return repr(self)



import fnmatch
import logging
import os
import warnings
from glob import glob
from pathlib import Path

import iris
import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom import const
from pyaerocom.exceptions import (
    DataCoverageError,
    DataQueryError,
    DataSourceError,
    FileConventionError,
    TemporalResolutionError,
    VariableDefinitionError,
    VarNotAvailableError,
)
from pyaerocom.griddeddata import GriddedData
from pyaerocom.units.datetime import get_highest_resolution, sort_ts_types, to_pandas_timestamp
from pyaerocom.helpers import isnumeric
from pyaerocom.io import AerocomBrowser
from pyaerocom.io.aux_components_fun import (
    calc_concNhno3_from_vmr,
    calc_concNnh3_from_vmr,
    calc_concNnh4,
    calc_concNno3pm10,
    calc_concNno3pm25,
    calc_concno3pm10,
    calc_concno3pm25,
    calc_concNtnh,
    calc_concNtno3,
    calc_sspm25,
)
from pyaerocom.io.aux_read_cubes import (
    add_cubes,
    compute_angstrom_coeff_cubes,
    divide_cubes,
    mmr_from_vmr,
    multiply_cubes,
    subtract_cubes,
)
from pyaerocom.io.file_conventions import FileConventionRead
from pyaerocom.io.gridded_reader import GriddedReader
from pyaerocom.io.helpers import add_file_to_log
from pyaerocom.io.iris_io import concatenate_iris_cubes, load_cubes_custom
from pyaerocom.metastandards import AerocomDataID
from pyaerocom.units.datetime import TsType
from pyaerocom.variable import Variable

logger = logging.getLogger(__name__)


class ReadGridded(GriddedReader):
    """Class for reading gridded files using AeroCom file conventions

    Attributes
    ----------
    data_id : str
        string ID for model or obsdata network (see e.g. Aerocom interface map
        plots lower left corner)
    data : GriddedData
        imported data object
    data_dir : str
        directory containing result files for this model
    start : pandas.Timestamp
        start time for data import
    stop : pandas.Timestamp
        stop time for data import
    file_convention : FileConventionRead
        class specifying details of the file naming convention for the model
    files : list
        list containing all filenames that were found.
        Filled, e.g. in :func:`ReadGridded.get_model_files`
    from_files : list
        List of all netCDF files that were used to concatenate the current
        data cube (i.e. that can be based on certain matching settings such as
        var_name or time interval).
    ts_types : list
        list of all sampling frequencies (e.g. hourly, daily, monthly) that
        were inferred from filenames (based on Aerocom file naming convention)
        of all files that were found
    vars : list
        list containing all variable names (e.g. od550aer) that were inferred
        from filenames based on Aerocom model file naming convention
    years : list
        list of available years as inferred from the filenames in the data
        directory.

    Parameters
    ----------
    data_id : str
        string ID of model (e.g. "AATSR_SU_v4.3","CAM5.3-Oslo_CTRL2016")
    data_dir : str, optional
        directory containing data files. If provided, only this directory is
        considered for data files, else the input `data_id` is used to search
        for the corresponding directory.
    file_convention : str
        string ID specifying the file convention of this model (cf.
        installation file `file_conventions.ini <https://github.com/metno/
        pyaerocom/blob/master/pyaerocom/data/file_conventions.ini>`__)
    init : bool
        if True, the model directory is searched (:func:`search_data_dir`) on
        instantiation and if it is found, all valid files for this model are
        searched using :func:`search_all_files`.

    """

    CONSTRAINT_OPERATORS = {
        "==": np.equal,
        "!=": np.not_equal,
        "<": np.less,
        "<=": np.less_equal,
        ">": np.greater,
        ">=": np.greater_equal,
    }

    AUX_REQUIRES = {
        "ang4487aer": ("od440aer", "od870aer"),
        "angabs4487aer": ("abs440aer", "abs870aer"),
        "od550gt1aer": ("od550aer", "od550lt1aer"),
        "wetoa": ("wetpoa", "wetsoa"),
        "dryoa": ("drypoa", "drysoa"),
        "conc*": ("mmr*", "rho"),
        "sc550dryaer": ("ec550dryaer", "ac550dryaer"),
        "mmr*": ("vmr*",),
        "rho": ("ts", "ps"),
        "concox": ("concno2", "conco3"),
        "vmrox": ("vmrno2", "vmro3"),
        "fmf550aer": ("od550lt1aer", "od550aer"),
        "concno3": ("concno3c", "concno3f"),
        "concprcpoxn": ("wetoxn", "pr"),
        "concprcpoxs": ("wetoxs", "pr"),
        "concprcprdn": ("wetrdn", "pr"),
        "concsspm10": ("concss25", "concsscoarse"),
        "concsspm25": ("concss25", "concsscoarse"),
        "concno3pm10": ("concno3f", "concno3c"),
        "concno3pm25": ("concno3f", "concno3c"),
        "concNno3pm10": ("concno3f", "concno3c"),
        "concNno3pm25": ("concno3f", "concno3c"),
        "concNhno3": ("vmrhno3",),
        "concNtno3": ("concno3f", "concno3c", "vmrhno3"),
        "concNnh3": ("vmrnh3",),
        "concNnh4": ("concnh4",),
        "concNtnh": ("concnh4", "vmrnh3"),
    }

    AUX_ALT_VARS = {
        "od440aer": ["od443aer"],
        "od870aer": ["od865aer"],
        "ac550dryaer": ["ac550aer"],
    }

    AUX_FUNS = {
        "ang4487aer": compute_angstrom_coeff_cubes,
        "angabs4487aer": compute_angstrom_coeff_cubes,
        "od550gt1aer": subtract_cubes,
        "wetoa": add_cubes,
        "dryoa": add_cubes,
        "sc550dryaer": subtract_cubes,
        "conc*": multiply_cubes,
        "mmr*": mmr_from_vmr,
        "concox": add_cubes,
        "vmrox": add_cubes,
        "fmf550aer": divide_cubes,
        "concno3": add_cubes,
        "concsspm10": add_cubes,
        "concsspm25": calc_sspm25,
        "concno3pm10": calc_concno3pm10,
        "concno3pm25": calc_concno3pm25,
        "concNno3pm10": calc_concNno3pm10,
        "concNno3pm25": calc_concNno3pm25,
        "concNhno3": calc_concNhno3_from_vmr,
        "concNtno3": calc_concNtno3,
        "concNnh3": calc_concNnh3_from_vmr,
        "concNnh4": calc_concNnh4,
        "concNtnh": calc_concNtnh,
        # 'mec550*'      :    divide_cubes,
        # 'tau*'         :    lifetime_from_load_and_dep
    }

    #: Additional arguments passed to computation methods for auxiliary data
    #: This is optional and defined per-variable like in AUX_FUNS
    AUX_ADD_ARGS = {
        "concprcpoxn": dict(
            ts_type="daily", prlim=0.1e-3, prlim_units="m d-1", prlim_set_under=np.nan
        ),
        "concprcpoxs": dict(
            ts_type="daily", prlim=0.1e-3, prlim_units="m d-1", prlim_set_under=np.nan
        ),
        "concprcprdn": dict(
            ts_type="daily", prlim=0.1e-3, prlim_units="m d-1", prlim_set_under=np.nan
        ),
    }

    _data_dir = ""

    VERT_ALT = {"Surface": "ModelLevel"}

    def __init__(self, data_id=None, data_dir=None, file_convention="aerocom3"):
        self._data_dir = None

        #: data_id of gridded dataset
        self._data_id = data_id

        # file naming convention. Default is aerocom3 file convention, change
        # using self.file_convention.import_default("aerocom2"). Is
        # automatically updated in class ReadGridded
        self.file_convention = FileConventionRead(file_convention)

        self.file_info = None

        #: List of unique Aerocom variable names that were identified from
        #: the filenames in the data directory
        self._vars_2d = []
        self._vars_3d = []

        #: This object can be used to
        self.browser = AerocomBrowser()

        # these can be filled using method add_aux_compute and they will not
        # affect global settings of the reader class
        self._aux_requires = {}
        self._aux_funs = {}

        #: quick access to variables that can be computed (will be filled and
        #: handles automatically)
        self._aux_avail = {}

        self.ignore_vert_code = False
        if data_dir is not None:
            self.data_dir = data_dir
        elif data_id:
            self.search_data_dir()
        if self.data_dir is not None:
            try:
                self.search_all_files()
            except DataCoverageError as e:
                logger.warning(repr(e))

    @property
    def data_id(self) -> str:
        """
        Data ID of dataset
        """
        return self._data_id

    @data_id.setter
    def data_id(self, val):
        if val is None:
            val = ""
        if not isinstance(val, str):
            raise ValueError("Invalid input for data_id, need str")
        self._data_id = val

    @property
    def data_dir(self) -> str:
        """
        Directory where data files are located
        """
        return self._data_dir

    @data_dir.setter
    def data_dir(self, val):
        if isinstance(val, Path):
            val = str(val)
        if not isinstance(val, str) or not os.path.isdir(val):
            raise FileNotFoundError(f"Input data directory {val} does not exist")
        self._data_dir = val
        self.reinit()

    @property
    def years_avail(self) -> list:
        """
        Years available in dataset
        """
        if self.file_info is None:
            self.search_all_files()
        return sorted(self.file_info.year.unique())

    @property
    def experiments(self) -> list:
        """
        List of all experiments that are available in this dataset
        """
        if self.file_info is None:
            self.search_all_files()
        return sorted(self.file_info.experiment.unique())

    @property
    def files(self) -> list:
        """
        List of data files
        """
        if self.file_info is None:
            self.search_all_files()
        return [os.path.join(self.data_dir, x) for x in sorted(self.file_info.filename.values)]

    @property
    def ts_types(self):
        """Available frequencies"""
        return self.file_info.ts_type.unique()

    @property
    def vars_filename(self):
        return sorted(self._vars_2d + self._vars_3d)

    @property
    def vars(self):
        from pyaerocom.exceptions import DeprecationError

        raise DeprecationError(
            "Attribute vars is deprecated in ReadGridded. Please use vars_filename instead"
        )

    @property
    def vars_provided(self):
        """Variables provided by this dataset"""
        return self._get_vars_provided()

    @property
    def file_type(self):
        """File type of data files"""
        return const.GRID_IO.FILE_TYPE

    @property
    def TS_TYPES(self):
        """List with valid filename encryptions specifying temporal resolution

        Update 7.11.2019: not in use anymore due to improved handling of
        all possible frequencies now using TsType class.
        """
        return const.GRID_IO.TS_TYPES

    @property
    def start(self):
        """First available year in the dataset (inferred from filenames)

        Note
        ----
        This is not variable or ts_type specific, so it is not necessarily
        given that data from this year is available for all variables in
        :attr:`vars` or all frequencies liste in :attr:`ts_types`
        """
        if len(self.years_avail) == 0:
            raise AttributeError(
                "No information about available years accessible"
                "please run method search_all_files first"
            )
        yr = sorted(self.years_avail)[0]
        if yr == 9999:
            yr = 2222
        return to_pandas_timestamp(yr)

    @property
    def stop(self):
        """Last available year in the dataset (inferred from filenames)

        Note
        ----
        This is not variable or ts_type specific, so it is not necessarily
        given that data from this year is available for all variables in
        :attr:`vars` or all frequencies liste in :attr:`ts_types`
        """
        if len(self.years_avail) == 0:
            raise AttributeError(
                "No information about available years accessible"
                "please run method search_all_files first"
            )
        years = sorted(self.years_avail)
        year = years[-1]

        if year == 9999:
            logger.warning(
                "Data contains climatology. Will be ignored as stop time, using last year"
            )
            if len(years) == 1:
                year = 2222
            else:
                year = years[-2]

        return to_pandas_timestamp(f"{year}-12-31 23:59:59")

    def reinit(self):
        """Reinit everything that is loaded specific to data_dir"""
        self.file_info = None
        self._vars_2d = []
        self._vars_3d = []

    def _get_vars_provided(self):
        """(Private method) get all variables provided"""
        _vars = []
        _vars.extend(self.vars_filename)

        for aux_var in self.AUX_REQUIRES:
            if "*" in aux_var:
                continue
            if aux_var not in _vars and self.check_compute_var(aux_var):
                _vars.append(aux_var)
        for aux_var in self._aux_requires:
            if "*" in aux_var:
                continue
            if aux_var not in _vars and self.check_compute_var(aux_var):
                _vars.append(aux_var)

        # also add standard names of 3D variables if not already in list
        for var in self._vars_3d:
            var = var.lower().replace("3d", "")
            if var not in _vars:
                _vars.append(var)
        return _vars

    def _check_aux_compute_access(self, var_name):
        """Check if input var_name can be computed

        Note
        ----
        Does not check variable families (e.g. conc*) which is done via
        :func:`_check_var_match_pattern`. Only checks entries in
        :attr:`AUX_REQUIRES` and :attr:`aux_requires`.

        Parameters
        ----------
        var_name : str
            input variable name that is supposed to be checked

        Returns
        -------
        bool
            True, if variable can be computed, else False
        """
        if var_name in self._aux_requires and var_name in self._aux_funs:
            vars_req = self._aux_requires[var_name]
            fun = self._aux_funs[var_name]
        elif var_name in self.AUX_REQUIRES and var_name in self.AUX_FUNS:
            vars_req = self.AUX_REQUIRES[var_name]
            fun = self.AUX_FUNS[var_name]
        else:
            return False
        vars_to_read = []
        for i, var in enumerate(vars_req):
            try:
                # the variable might be updated in _check_var_avail
                vars_to_read.append(self._check_var_avail(var))
            except (VarNotAvailableError, VariableDefinitionError):
                return False
        if not len(vars_to_read) == len(vars_req):
            return False
        self._aux_avail[var_name] = (vars_to_read, fun)
        return True

    def _check_var_match_pattern(self, var_name):
        """Check if input variable can be accessed via auxiliary variable family

        E.g. if var_name is concpm10 and mmrpm10 and rho are available

        Parameters
        ----------
        var_name : str
            variable that is supposed to be read

        Returns
        -------
        bool
            True if variable can be read, else False
        """
        vars_found = []
        for pattern in self.registered_var_patterns:
            if fnmatch.fnmatch(var_name, pattern):
                vars_required = self.AUX_REQUIRES[pattern]
                for addvar in vars_required:
                    if "*" not in addvar:
                        vars_found.append(addvar)
                    else:
                        _addvar = var_name
                        spl1 = pattern.split("*")
                        spl2 = addvar.split("*")
                        if len(spl1) != len(spl2):
                            raise AttributeError(
                                "variable patterns in "
                                "AUX_REQUIRES and corresponding "
                                "values (with * in name) need "
                                "to have the same number of "
                                "wildcard delimiters"
                            )
                        for i, substr in enumerate(spl1):
                            if bool(substr):
                                _addvar = _addvar.replace(substr, spl2[i])
                        vars_found.append(_addvar)

                if len(vars_found) == len(vars_required):
                    all_ok = True
                    vars_to_read = []
                    for var in vars_found:
                        try:
                            vars_to_read.append(self._check_var_avail(var))
                        except (VarNotAvailableError, VariableDefinitionError):
                            all_ok = False
                            break

                    if all_ok:
                        fun = self.AUX_FUNS[pattern]
                        self.add_aux_compute(var_name, vars_required=vars_to_read, fun=fun)
                        self._aux_avail[var_name] = (vars_to_read, fun)
                        return True
        return False

    def _get_aux_vars_and_fun(self, var_to_compute):
        """Helper that searches auxiliary variables for computation of input var

        Parameters
        ----------
        var_to_compute : str
            one of the auxiliary variables that is supported by this interface
            (cf. :attr:`AUX_REQUIRES`)

        Raises
        ------
        VarNotAvailableError
            if one of the required variables for computation is not available
            in the data

        Returns
        -------
        list
            list of variables that are used as input for computation method
            of input variable
        callable
            function that is used to compute input variable
        """
        if var_to_compute not in self._aux_avail:
            if not self.check_compute_var(var_to_compute):
                raise VarNotAvailableError(f"Variable {var_to_compute} cannot be computed")
        return self._aux_avail[var_to_compute]

    def check_compute_var(self, var_name):
        """Check if variable name belongs to family that can be computed

        For instance, if input var_name is `concdust` this method will check
        :attr:`AUX_REQUIRES` to see if there is a variable family pattern
        (`conc*`) defined that specifies how to compute these variables. If
        a match is found, the required variables and computation method is
        added via :func:`add_aux_compute`.

        Parameters
        -----------
        var_name : str
            variable name to be checked

        Returns
        -------
        bool
            True if match is found, else False
        """

        if "*" in var_name:
            raise VariableDefinitionError(f"Invalid variable name {var_name}. Must not contain *")
        if var_name in self._aux_avail:
            return True
        elif self._check_aux_compute_access(var_name):
            return True
        return self._check_var_match_pattern(var_name)

    def _check_var_avail(self, var):
        if var in self.vars_filename:
            return var

        if var in self.AUX_ALT_VARS:
            for alt_var in self.AUX_ALT_VARS[var]:
                if alt_var in self.vars_filename:
                    return alt_var

        v = const.VARS[var]

        for alias in v.aliases:
            if alias in self.vars_filename:
                return alias
        raise VarNotAvailableError(f"Var {var} is not available in data...")

    def has_var(self, var_name):
        """Check if variable is available

        Parameters
        ----------
        var_name : str
            variable to be checked

        Returns
        -------
        bool
        """
        # vars_provided includes variables that can be read and variables that
        # can be computed. It does not consider variable families that may be
        # able to be computed or alias matches
        avail = self.vars_provided
        if var_name in avail:
            return True
        try:
            var = const.VARS[var_name]
        except VariableDefinitionError as e:
            logger.warning(repr(e))
            return False

        if self.check_compute_var(var_name):
            return True

        for alias in var.aliases:
            if alias in avail:
                return True

        if var.is_alias and var.var_name_aerocom in avail:
            return True

        return False

    def _get_years_to_load(self, start=None, stop=None):
        """Array containing year numbers that are supposed to be loaded

        Returns
        -------
        ndarray
            all years to be loaded
        """
        start_provided = False
        if start is None:
            start = self.start
            if start.year == 2222:
                return np.array([9999])
        else:
            start_provided = True
            if isinstance(start, int) and start == 9999:
                return np.array([9999])
            start = to_pandas_timestamp(start)

        if stop is None:
            if start_provided:
                stop = to_pandas_timestamp(start)
            else:
                stop = self.stop
        else:
            stop = to_pandas_timestamp(stop)
            stop -= np.timedelta64(
                1, "s"
            )  # subtract one second to end up at the end of previous year
        if const.MIN_YEAR > start.year:
            logger.warning(
                f"First available year {start} of data {self.data_id} is smaller "
                f"than supported first year {const.MIN_YEAR}."
            )
            start = const.MIN_YEAR
        if const.MAX_YEAR < stop.year:
            raise ValueError(
                f"Last available year {stop} of data {self.data_id} is larger "
                f"than supported last year {const.MAX_YEAR}."
            )
            stop = const.MAX_YEAR

        if start and stop:
            return np.arange(start.year, stop.year + 1, 1)

        if not self.years_avail:
            raise AttributeError(
                "No information available for available "
                "years. Please run method "
                "search_all_files first"
            )
        return np.array(self.years_avail)

    def search_data_dir(self):
        """Search data directory based on model ID

        Wrapper for method :func:`search_data_dir_aerocom`

        Returns
        -------
        str
            data directory

        Raises
        ------
        IOError
            if directory cannot be found
        """
        _dir = self.browser.find_data_dir(self.data_id)
        self.data_dir = _dir
        return _dir

    @staticmethod
    def _eval_data_id(data_id):
        """
        Extract meta information from data_id

        Parameters
        ----------
        data_id : str
            string specifying model and other relevant information. In the
            best case following AeroCom Phase III conventions, that is,
            following the template specified in :class:`AerocomDataID`.

        Returns
        -------
        vals : list
            list containing successfully extracted information from data_id
        """

        vals = AerocomDataID(data_id).values
        return vals

    def _update_file_convention(self, files):
        """Update current file convention based on input files

        Loops over all files in input list and as updates the file convention
        based on the first file in list that matches one of the registered
        conventions.

        Updates class :attr:`file_convention`

        Raises
        ------
        FileNotFoundError
            if none of the input files matches a registered convention.
        """
        for file in files:
            try:
                self.file_convention.from_file(os.path.basename(file))
                return
            except Exception:
                pass

        raise FileNotFoundError(
            f"None of the available files in {self.data_dir} matches a "
            f"registered pyaerocom file convention"
        )

    def _evaluate_fileinfo(self, files):
        result = []
        _vars_temp = []
        _vars_temp_3d = []

        for _file in files:
            # TODO: resolve this in a more general way...
            if "ModelLevelAtStations" in _file:
                logger.info(f"Ignoring file {_file}")
                continue
            try:
                info = self.file_convention.get_info_from_file(_file)
                if not self.data_id:
                    self.data_id = info["data_id"]
                var_name = info["var_name"]
                _is_3d = False
                if is_3d(var_name):
                    _vars_temp_3d.append(var_name)
                    _is_3d = True
                else:
                    _vars_temp.append(var_name)

                if not TsType.valid(info["ts_type"]):  # in self.TS_TYPES:
                    raise TemporalResolutionError(f"Invalid frequency {info['ts_type']}")

                (model, meteo, experiment, pert) = self._eval_data_id(info["data_id"])
                result.append(
                    [
                        var_name,
                        info["year"],
                        info["ts_type"],
                        info["vert_code"],
                        self.data_id,
                        model,
                        meteo,
                        experiment,
                        pert,
                        info["is_at_stations"],
                        _is_3d,
                        os.path.basename(_file),
                    ]
                )

            except (FileConventionError, DataSourceError, TemporalResolutionError) as e:
                msg = f"Failed to import file\n{_file}\nModel: {self.data_id}\nError: {e}"
                logger.warning(msg)
                if const.WRITE_FILEIO_ERR_LOG:
                    add_file_to_log(_file, msg)

        if len(_vars_temp + _vars_temp_3d) == 0:
            raise AttributeError("Failed to extract information from filenames")

        self._vars_2d = sorted(set(_vars_temp))
        self._vars_3d = sorted(set(_vars_temp_3d))
        return result

    def _fileinfo_to_dataframe(self, result):
        header = [
            "var_name",
            "year",
            "ts_type",
            "vert_code",
            "data_id",
            "name",
            "meteo",
            "experiment",
            "perturbation",
            "is_at_stations",
            "3D",
            "filename",
        ]

        df = pd.DataFrame(result, columns=header)
        df.sort_values(
            [
                "var_name",
                "year",
                "ts_type",
                "data_id",
                "name",
                "meteo",
                "experiment",
                "perturbation",
                "is_at_stations",
                "3D",
            ],
            inplace=True,
        )

        uv = df.vert_code.unique()
        if len(uv) == 1 and uv[0] == "":
            self.ignore_vert_code = True
        return df

    def search_all_files(self, update_file_convention=True):
        """Search all valid model files for this model

        This method browses the data directory and finds all valid files, that
        is, file that are named according to one of the aerocom file naming
        conventions. The file list is stored in :attr:`files`.

        Note
        ----
        It is presumed, that naming conventions of files in
        the data directory are not mixed but all correspond to either of the
        conventions defined in

        Parameters
        ----------
        update_file_convention : bool
            if True, the first file in `data_dir` is used to identify the
            file naming convention (cf. :class:`FileConventionRead`)

        Raises
        ------
        DataCoverageError
            if no valid files could be found
        """
        if self.data_dir is None:
            raise AttributeError("please set data_dir first")

        # get all files with correct ending
        files = glob(f"{self.data_dir}/*{self.file_type}")
        if len(files) == 0:
            logger.warning(
                f"No files of type {self.file_type} could be found in current "
                f"data directory (data_dir={os.path.abspath(self.data_dir)}"
            )
            return

        if update_file_convention:
            # Check if the found file has a naming according the aerocom conventions
            # and set the convention for all files (maybe this need to be
            # updated in case there can be more than one file naming convention
            # within one model directory)
            try:
                self._update_file_convention(files)
            except FileNotFoundError as e:
                logger.warning(repr(e))
                return

        result = self._evaluate_fileinfo(files)
        df = self._fileinfo_to_dataframe(result)
        self.file_info = df

        if len(df) == 0:
            raise DataCoverageError(f"No valid files could be found for {self.data_id}")

    def filter_files(
        self,
        var_name=None,
        ts_type=None,
        start=None,
        stop=None,
        experiment=None,
        vert_which=None,
        is_at_stations=False,
        df=None,
    ):
        """Filter file database

        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        ts_type : str
            string specifying temporal resolution (choose from
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised
            of the available resolutions is used
        start : Timestamp or str, optional
            start time of data import
        stop : Timestamp or str, optional
            stop time of data import
        experiment : str
            name of experiment (only relevant if this dataset contains more
            than one experiment)
        vert_which : str or dict, optional
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel) or dictionary containing var_name as key and vertical
            coded string as value, accordingly
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that
            match the query.

        """
        if df is None:
            df = self.file_info

        yrs = self._get_years_to_load(start, stop)
        year_cond = df.year.isin(yrs)

        if var_name is None:
            var_cond = df.var_name.isin(df.var_name.values)
        else:
            var_cond = df.var_name == var_name
        if vert_which is None:
            vert_cond = df.vert_code.isin(df.vert_code.values)
        else:
            vert_cond = df.vert_code == vert_which
        if ts_type is None:
            freq_cond = df.ts_type.isin(df.ts_type.values)
        else:
            freq_cond = df.ts_type == ts_type
        if experiment is None:
            exp_cond = df.experiment.isin(df.experiment.values)
        else:
            exp_cond = df.experiment == experiment

        return df.loc[
            (var_cond)
            & (year_cond)
            & (freq_cond)
            & (exp_cond)
            & (vert_cond)
            & (df.is_at_stations == is_at_stations)
        ]

    def _infer_ts_type(self, df, ts_type, flex_ts_type, prefer_longer):
        ts_types = df.ts_type.unique()

        if len(ts_types) == 1:
            # only one frequency available
            if flex_ts_type or ts_type is None or ts_types[0] == ts_type:
                # all good
                return ts_types[0]
            raise DataCoverageError(f"No files could be found for ts_type {ts_type}")
        highest_avail = get_highest_resolution(*ts_types)
        # there is more than one frequency available -> decision making
        # gets more complicated
        if not flex_ts_type:
            if ts_type is None:
                return highest_avail
            elif ts_type in ts_types:
                return ts_type
            raise DataCoverageError("Failed to infer ts_type")

        # ts_type is flexible
        if ts_type is None:
            # initiate with highest available
            ts_type = highest_avail

        if not prefer_longer:
            return ts_type

        # ts_type is flexible and user prefers the longer period over
        # higher resolution
        ts_type = ts_types[0]
        subset = self.filter_files(ts_type=ts_type, df=df)
        for _ts_type in ts_types[1:]:
            _subset = self.filter_files(ts_type=_ts_type, df=df)
            if len(_subset) > len(subset):
                subset = _subset
                ts_type = _ts_type
        return ts_type

    def filter_query(
        self,
        var_name,
        ts_type=None,
        start=None,
        stop=None,
        experiment=None,
        vert_which=None,
        is_at_stations=False,
        flex_ts_type=True,
        prefer_longer=False,
    ):
        """Filter files for read query based on input specs

        Parameters
        ----------

        Returns
        -------
        DataFrame
            dataframe containing filtered dataset
        """
        if var_name not in self.file_info.var_name.values:
            raise DataCoverageError(
                f"Variable {var_name} is not available in dataset {self.data_id}"
            )

        subset = self.filter_files(
            var_name=var_name,
            ts_type=None,  # disregard ts_type in 1. iteration
            start=start,
            stop=stop,
            experiment=experiment,
            vert_which=vert_which,
            is_at_stations=is_at_stations,
        )
        if len(subset) == 0:
            if vert_which in self.VERT_ALT:
                vc = self.VERT_ALT[vert_which]
                logger.warning(
                    f"No files could be found for var {var_name} and "
                    f"vert_which {vert_which} in {self.data_id}. "
                    f"Trying to find alternative options"
                )
                return self.filter_query(
                    var_name,
                    ts_type,
                    start,
                    stop,
                    experiment,
                    vert_which=vc,
                    is_at_stations=is_at_stations,
                    flex_ts_type=flex_ts_type,
                    prefer_longer=prefer_longer,
                )
            raise DataCoverageError("No files could be found")
        ts_type = self._infer_ts_type(subset, ts_type, flex_ts_type, prefer_longer)
        subset = self.filter_files(ts_type=ts_type, start=start, stop=stop, df=subset)
        if len(subset) == len(subset.year.unique()):
            return subset

        # File request could not be resolved such that every year only occurs
        # once
        msg = ""
        exps = subset.experiment.unique()
        verts = subset.vert_code.unique()

        if len(exps) > 1:
            msg += f"Found multiple experiments. Choose from: {exps}"
        if len(verts) > 1:
            dvc = const.VARS[var_name].get_default_vert_code()
            if dvc is not None and dvc in verts:
                return self.filter_query(
                    var_name,
                    ts_type,
                    start,
                    stop,
                    experiment,
                    vert_which=dvc,
                    is_at_stations=is_at_stations,
                    flex_ts_type=flex_ts_type,
                    prefer_longer=prefer_longer,
                )

            if msg:
                msg += "; "
            msg += f"Found multiple vertical codes. Choose from: {verts}"
        raise DataQueryError(
            f"Failed to uniquely identify data files for input query. Reason: {msg}"
        )

    def get_files(
        self,
        var_name,
        ts_type=None,
        start=None,
        stop=None,
        experiment=None,
        vert_which=None,
        is_at_stations=False,
        flex_ts_type=True,
        prefer_longer=False,
    ):
        """Get data files based on input specs"""
        subset = self.filter_query(
            var_name,
            ts_type,
            start,
            stop,
            experiment,
            vert_which,
            is_at_stations,
            flex_ts_type,
            prefer_longer,
        )

        return self._generate_file_paths(subset)

    def _generate_file_paths(self, df=None):
        if df is None:
            df = self.file_info
        return sorted(os.path.join(self.data_dir, x) for x in df.filename.values)

    def get_var_info_from_files(self) -> dict:
        """Creates dictionary that contains variable specific meta information

        Returns
        -------
        dict
            dictionary where keys are available variables and values (for each
            variable) contain information about available ts_types, years, etc.
        """
        result = {}
        for file in self.files:
            finfo = self.file_convention.get_info_from_file(file)
            var_name = finfo["var_name"]
            if var_name not in result:
                result[var_name] = var_info = {}
                for key in finfo:
                    if key != "var_name":
                        var_info[key] = []
            else:
                var_info = result[var_name]
            for key, val in finfo.items():
                if key == "var_name":
                    continue
                if val is not None and val not in var_info[key]:
                    var_info[key].append(val)
        # now check auxiliary variables
        for var_to_compute in self.AUX_REQUIRES:
            if var_to_compute in result:
                continue
            try:
                vars_to_read = self._get_aux_vars_and_fun(var_to_compute)[0]
            except VarNotAvailableError:
                pass
            except VariableDefinitionError:
                pass
            else:
                # init result info dict for aux variable
                result[var_to_compute] = var_info = {}
                first = result[vars_to_read[0]]
                # init with results from first required variable
                var_info.update(**first)
                if len(vars_to_read) > 1:
                    for info_other in vars_to_read[1:]:
                        other = result[info_other]
                        for key, info in var_info.items():
                            # compute match with other variable
                            var_info[key] = list(np.intersect1d(info, other[key]))
                var_info["aux_vars"] = vars_to_read

        return result

    def update(self, **kwargs):
        """Update one or more valid parameters

        Parameters
        ----------
        **kwargs
            keyword args that will be used to update (overwrite) valid class
            attributes such as `data, data_dir, files`
        """
        for k, v in kwargs.items():
            if k in self.__dict__:
                logger.info(
                    f"Updating {k} in ModelImportResult for model {self.data_id}. New value: {v}"
                )
                self.__dict__[k] = v
            else:
                logger.info(f"Ignoring key {k} in ModelImportResult.update()")

    def concatenate_cubes(self, cubes):
        """Concatenate list of cubes into one cube

        Parameters
        ----------
        CubeList
            list of individual cubes

        Returns
        -------
        Cube
            Single cube that contains concatenated cubes from input list

        Raises
        ------
        iris.exceptions.ConcatenateError
            if concatenation of all cubes failed
        """
        return concatenate_iris_cubes(cubes, error_on_mismatch=True)

    def compute_var(
        self,
        var_name,
        start=None,
        stop=None,
        ts_type=None,
        experiment=None,
        vert_which=None,
        flex_ts_type=True,
        prefer_longer=False,
        vars_to_read=None,
        aux_fun=None,
        try_convert_units=True,
        aux_add_args=None,
        rename_var=None,
        **kwargs,
    ):
        """Compute auxiliary variable

        Like :func:`read_var` but for auxiliary variables
        (cf. AUX_REQUIRES)

        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        start : Timestamp or str, optional
            start time of data import (if valid input, then the current
            :attr:`start` will be overwritten)
        stop : Timestamp or str, optional
            stop time of data import
        ts_type : str
            string specifying temporal resolution (choose from
            hourly, 3hourly, daily, monthly). If None, prioritised
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more
            than one experiment)
        vert_which : str
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel)
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that
            match the query.
        try_convert_units : bool
            if True, units of GriddedData objects are attempted to be converted
            to AeroCom default. This applies both to the GriddedData objects
            being read for computation as well as the variable computed from
            the forme objects. This is, for instance, useful when computing
            concentration in precipitation from wet deposition and precipitation
            amount.
        rename_var : str
            if this is set, the `var_name` attribute of the output
            `GriddedData` object will be updated accordingly.
        **kwargs
            additional keyword args passed to :func:`_load_var`

        Returns
        -------
        GriddedData
            loaded data object
        """
        if vars_to_read is not None:
            self.add_aux_compute(var_name, vars_to_read, aux_fun)
        vars_to_read, aux_fun = self._get_aux_vars_and_fun(var_name)
        if aux_add_args is None:
            aux_add_args = {}
        data = []
        # all variables that are required need to be in the same temporal
        # resolution
        try:
            ts_type = self.find_common_ts_type(
                vars_to_read,
                start,
                stop,
                ts_type,
                experiment,
                vert_which=vert_which,
                flex_ts_type=flex_ts_type,
            )
        except DataCoverageError:
            if vert_which not in self.VERT_ALT:
                raise
            vert_which = self.VERT_ALT[vert_which]

            ts_type = self.find_common_ts_type(
                vars_to_read,
                start,
                stop,
                ts_type,
                experiment,
                vert_which=vert_which,
                flex_ts_type=flex_ts_type,
            )
        for var in vars_to_read:
            aux_data = self._load_var(
                var_name=var,
                ts_type=ts_type,
                start=start,
                stop=stop,
                experiment=experiment,
                vert_which=vert_which,
                flex_ts_type=flex_ts_type,
                prefer_longer=prefer_longer,
                try_convert_units=try_convert_units,
                rename_var=None,
                **kwargs,
            )
            data.append(aux_data)

        if var_name in self.AUX_ADD_ARGS:
            for key, val in self.AUX_ADD_ARGS[var_name].items():
                if key not in aux_add_args:
                    aux_add_args[key] = val

        if len(aux_add_args) > 0:
            cube = aux_fun(*data, **aux_add_args)
        else:
            cube = aux_fun(*data)

        cube.var_name = var_name

        data = GriddedData(
            cube,
            data_id=self.data_id,
            computed=True,
            convert_unit_on_init=try_convert_units,
            **kwargs,
        )

        data.reader = self
        if rename_var is not None:
            data.var_name = rename_var
        return data

    def find_common_ts_type(
        self,
        vars_to_read,
        start=None,
        stop=None,
        ts_type=None,
        experiment=None,
        vert_which=None,
        flex_ts_type=True,
    ):
        """Find common ts_type for list of variables to be read

        Parameters
        ----------
        vars_to_read : list
            list of variables that is supposed to be read
        start : Timestamp or str, optional
            start time of data import (if valid input, then the current
            start will be overwritten)
        stop : Timestamp or str, optional
            stop time of data import (if valid input, then the current
            :attr:`start` will be overwritten)
        ts_type : str
            string specifying temporal resolution (choose from
            hourly, 3hourly, daily, monthly). If None, prioritised
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more
            than one experiment)
        vert_which : str
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel)
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case
            the input ts_type is not available for this variable

        Returns
        -------
        str
            common ts_type for input variable

        Raises
        ------
        DataCoverageError
            if no match can be found
        """
        if isinstance(vars_to_read, str):
            vars_to_read = [vars_to_read]

        common = self.filter_files(
            var_name=vars_to_read[0],
            start=start,
            stop=stop,
            experiment=experiment,
            vert_which=vert_which,
        ).ts_type.unique()
        if len(common) == 0:
            raise DataCoverageError(
                f"Could not find any file matches for query and variable {vars_to_read[0]}"
            )
        for var in vars_to_read[1:]:
            _tt = self.filter_files(
                var_name=var, start=start, stop=stop, experiment=experiment, vert_which=vert_which
            )
            common = np.intersect1d(common, _tt.ts_type.unique())

        if len(common) == 0:
            raise DataCoverageError(f"Could not find common ts_type for variables {vars_to_read}")
        elif len(common) == 1:
            if ts_type is None or flex_ts_type:
                return common[0]
            elif ts_type == common[0]:
                return ts_type
            raise DataCoverageError(
                f"Could not find files with ts_type={ts_type} for all input variables: {vars_to_read}"
            )
        if ts_type is not None:
            if ts_type in common:
                return ts_type

        if not flex_ts_type:
            raise DataCoverageError(
                f"Could not find files with ts_type={ts_type} for "
                f"all input variables: {vars_to_read}"
            )

        # NOTE: Changed by jgliss on 7.11.2019 for more flexibility
        # common_sorted = [x for x in const.GRID_IO.TS_TYPES if x in common]
        common_sorted = sort_ts_types(common)
        return common_sorted[0]

    def add_aux_compute(self, var_name, vars_required, fun):
        """Register new variable to be computed

        Parameters
        ----------
        var_name : str
            variable name to be computed
        vars_required : list
            list of variables to read, that are required to compute `var_name`
        fun : callable
            function that takes a list of `GriddedData` objects as input and
            that are read using variable names specified by `vars_required`.
        """
        if isinstance(vars_required, str):
            vars_required = [vars_required]
        if not isinstance(vars_required, list):
            raise ValueError(
                f"Invalid input for vars_required. Need str or list. Got: {vars_required}"
            )
        elif not callable(fun):
            raise ValueError("Invalid input for fun. Input is not a callable object")
        self._aux_requires[var_name] = vars_required
        self._aux_funs[var_name] = fun
        if not self._check_aux_compute_access(var_name):
            raise DataCoverageError(
                f"Failed to confirm access to auxiliary variable {var_name} from {vars_required}"
            )

    @property
    def registered_var_patterns(self):
        """
        List of string patterns for computation of variables

        The information is extracted from :attr:`AUX_REQUIRES`

        Returns
        -------
        list
            list of variable patterns

        """
        return [x for x in self.AUX_REQUIRES if "*" in x]

    def _get_var_to_read(self, var_name: str) -> str:
        """
        Get variable to read

        The logical order of inferring the variable to read is:

            1. Check if input variable name as is, is available in filename
            2. If not, check if an old version of that name is available
            3. If not, check if an alias name of input name is available
            4. If not, check if input variable is an alias and if AeroCom \
                variable name is available instead (e.g. input od550csaer \
                not available, but od550aer is available)

        Parameters
        ----------
        var_name : str
            name of variable that is supposed to be read

        Raises
        ------
        VarNotAvailableError
            if no match can be found
        VariableDefinitionError
            if input variable is not defined

        Returns
        -------
        str
            name of variable match found
        """

        if var_name in self.vars_filename:
            return var_name

        # e.g. user asks for od550aer but files contain only 3d var od5503daer
        # if not var_to_read in self.vars_filename:
        for var in self._vars_3d:
            if Variable(var).var_name == var_name:
                return var

        # get instance of Variable class
        var = const.VARS[var_name]

        # If the input variable has aliases, check if one of these aliases is
        # provided in the dataset
        for alias in var.aliases:
            if alias in self.vars_filename:
                logger.info(f"Did not find {var_name} field but {alias}. Using the latter instead")
                return alias

        # Finally, if still no match could be found, check if input variable
        # is an alias and see if the corresponding AeroCom variable name is
        # available in dataset
        if var.is_alias and var.var_name_aerocom in self.vars_filename:
            return var.var_name_aerocom

        raise VarNotAvailableError(f"Variable {var_name} could not be found")

    def _eval_vert_which_and_ts_type(self, var_name, vert_which, ts_type):
        if all(x == "" for x in self.file_info.vert_code.values):
            logger.info(
                f"Deactivating file search by vertical code for {self.data_id}, "
                f"since filenames do not include information about vertical code "
                f"(probably AeroCom 2 convention)"
            )
            vert_which = None

        if isinstance(vert_which, dict):
            try:
                vert_which = vert_which[var_name]
            except Exception:
                logger.info(
                    f"Setting vert_which to None, since input dict {vert_which} "
                    f"does not contain input variable {var_name}"
                )
                vert_which = None

        if isinstance(ts_type, dict):
            try:
                ts_type = ts_type[var_name]
            except Exception:
                logger.info(
                    f"Setting ts_type to None, since input dict {ts_type} "
                    f"does not contain specification variable to read {var_name}"
                )
                ts_type = None
        return vert_which, ts_type

    # TODO: add from_vars input arg for computation and corresponding method
    def read_var(
        self,
        var_name,
        start=None,
        stop=None,
        ts_type=None,
        experiment=None,
        vert_which=None,
        flex_ts_type=True,
        prefer_longer=False,
        aux_vars=None,
        aux_fun=None,
        constraints=None,
        try_convert_units=True,
        rename_var=None,
        **kwargs,
    ):
        """Read model data for a specific variable

        This method searches all valid files for a given variable and for a
        provided temporal resolution (e.g. *daily, monthly*), optionally
        within a certain time window, that may be specified on class
        instantiation or using the corresponding input parameters provided in
        this method.

        The individual NetCDF files for a given temporal period are loaded as
        instances of the :class:`iris.Cube` object and appended to an instance
        of the :class:`iris.cube.CubeList` object. The latter is then used to
        concatenate the individual cubes in time into a single instance of the
        :class:`pyaerocom.GriddedData` class. In order to ensure that this
        works, several things need to be ensured, which are listed in the
        following and which may be controlled within the global settings for
        NetCDF import using the attribute :attr:`GRID_IO` (instance of
        :class:`OnLoad`) in the default instance of the
        :class:`pyaerocom.config.Config` object accessible via
        ``pyaerocom.const``.

        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        start : Timestamp or str, optional
            start time of data import
        stop : Timestamp or str, optional
            stop time of data import
        ts_type : str
            string specifying temporal resolution (choose from
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more
            than one experiment)
        vert_which : str or dict, optional
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel) or dictionary containing var_name as key and vertical
            coded string as value, accordingly
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that
            match the query.
        aux_vars : list
            only relevant if `var_name` is not available for reading but needs
            to be computed: list of variables that are required to compute
            `var_name`
        aux_fun : callable
            only relevant if `var_name` is not available for reading but needs
            to be computed: custom method for computation (cf.
            :func:`add_aux_compute` for details)
        constraints : list, optional
            list of reading constraints (dict type). See
            :func:`check_constraint_valid` and :func:`apply_read_constraint`
            for details related to format of the individual constraints.
        try_convert_units : bool
            if True, then the unit of the variable data is checked against
            AeroCom default unit for that variable and if it deviates, it is
            attempted to be converted to the AeroCom default unit. Default is
            True.
        rename_var : str
            if this is set, the `var_name` attribute of the output
            `GriddedData` object will be updated accordingly.
        **kwargs
            additional keyword args parsed to :func:`_load_var`

        Returns
        -------
        GriddedData
            loaded data object

        Raises
        ------
        AttributeError
            if none of the ts_types identified from file names is valid
        VarNotAvailableError
            if specified ts_type is not supported
        """
        # user has provided input that specified how the input variable
        # is supposed to be computed. In this case, the variable will be
        # computed even if it is directly available in a file, i.e. it
        # could be read.
        if aux_vars is not None:
            self.add_aux_compute(var_name, aux_vars, aux_fun)

        vert_which, ts_type = self._eval_vert_which_and_ts_type(var_name, vert_which, ts_type)
        data = self._try_read_var(
            var_name,
            start,
            stop,
            ts_type,
            experiment,
            vert_which,
            flex_ts_type,
            prefer_longer,
            try_convert_units=try_convert_units,
            rename_var=rename_var,
            **kwargs,
        )

        if constraints is not None:
            if isinstance(constraints, dict):
                constraints = [constraints]
            for constraint in constraints:
                data = self.apply_read_constraint(
                    data,
                    constraint,
                    start=start,
                    stop=stop,
                    ts_type=ts_type,
                    experiment=experiment,
                    vert_which=vert_which,
                    flex_ts_type=flex_ts_type,
                    prefer_longer=prefer_longer,
                    **kwargs,
                )
        return data

    def check_constraint_valid(self, constraint):
        """
        Check if reading constraint is valid

        Parameters
        ----------
        constraint : dict
            reading constraint. Requires at lest entries for following keys:
            - operator (str): for valid operators see :attr:`CONSTRAINT_OPERATORS`
            - filter_val (float): value against which data is evaluated wrt to \
                operator

        Raises
        ------
        ValueError
            If constraint is invalid

        Returns
        -------
        None.

        """
        if not isinstance(constraint, dict):
            raise ValueError("Read constraint needs to be dict")
        elif "operator" not in constraint:
            raise ValueError(
                f"Constraint requires specification of operator. "
                f"Valid operators: {self.CONSTRAINT_OPERATORS}"
            )
        elif constraint["operator"] not in self.CONSTRAINT_OPERATORS:
            raise ValueError(
                f"Invalid constraint operator. Choose from: {self.CONSTRAINT_OPERATORS}"
            )
        elif "filter_val" not in constraint:
            raise ValueError("constraint needs specification of filter_val")
        elif not isnumeric(constraint["filter_val"]):
            raise ValueError("Need numerical filter value")

    def apply_read_constraint(self, data, constraint, **kwargs):
        """
        Filter a `GriddeData` object by value in another variable

        Note
        ----
        BETA version, that was hacked down in a rush to be able to apply
        AOD>0.1 threshold when reading AE.


        Parameters
        ----------
        data : GriddedData
            data object to which constraint is applied
        constraint : dict
            dictionary defining read constraint (see
            :func:`check_constraint_valid` for minimum requirement). If
            constraint contains key var_name (not mandatory), then the
            corresponding variable is attempted to be read and is used to
            evaluate constraint and the corresponding boolean mask is then
            applied to input `data`. Wherever this mask is True (i.e. constraint
            is met), the current value in input `data` will be replaced with
            `numpy.ma.masked` or, if specified, with entry `new_val` in input
            constraint dict.
        **kwargs : TYPE
            reading arguments in case additional variable data needs to be
            loaded, to determine filter mask (i.e. if `var_name` is specified
            in input constraint). Parse to :func:`read_var`.

        Raises
        ------
        ValueError
            If constraint is invalid (cf. :func:`check_constraint_valid` for
            details).

        Returns
        -------
        GriddedData
            modified data objects (all grid-points that met constraint
            are replaced with either `numpy.ma.masked` or with a
            value that can be specified via key `new_val` in input constraint).

        """
        self.check_constraint_valid(constraint)
        if "new_val" in constraint:
            new_val = constraint["new_val"]
        else:
            new_val = np.nan

        operator_fun = self.CONSTRAINT_OPERATORS[constraint["operator"]]

        if "var_name" in constraint:
            other_data = self.read_var(constraint["var_name"], **kwargs)
        else:
            other_data = data

        if not other_data.shape == data.shape:
            raise ValueError("Failed to apply filter. Shape mismatch")

        other_arr = other_data.to_xarray()
        arr = data.to_xarray()
        if not other_arr.dims == arr.dims:
            from pyaerocom.exceptions import DataDimensionError

            raise DataDimensionError("Mismatch in dimensions")
        for dim in arr.dims:
            same_vals = (arr[dim].values == other_arr[dim].values).all()
            if not same_vals:
                if dim == "time":
                    other_arr[dim] = arr[dim]
                else:
                    raise ValueError()

        # select all grid points where conition is fulfilled
        mask = operator_fun(other_arr, constraint["filter_val"])

        # set values to new_val where condition is fulfilled
        filtered = xr.where(mask, new_val, arr)

        # overwrite data in cube with the filtered data
        outcube = filtered.to_iris()
        outcube.var_name = data.var_name
        outcube.units = data.units
        outcube.attributes = data.cube.attributes
        outcube_dims = outcube.dim_coords
        for i, name in enumerate(data.dimcoord_names):
            _name = outcube_dims[i].name()
            outcube.remove_coord(_name)
            outcube.add_dim_coord(data.cube.coord(name), i)
        data.cube = outcube
        return data

    def _try_read_var(
        self,
        var_name,
        start,
        stop,
        ts_type,
        experiment,
        vert_which,
        flex_ts_type,
        prefer_longer,
        try_convert_units,
        rename_var,
        **kwargs,
    ):
        """Helper method used in :func:`read_var`

        See :func:`read_var` for description of input arguments.
        """
        if var_name in self._aux_requires and self.check_compute_var(var_name):
            return self.compute_var(
                var_name=var_name,
                start=start,
                stop=stop,
                ts_type=ts_type,
                experiment=experiment,
                vert_which=vert_which,
                flex_ts_type=flex_ts_type,
                prefer_longer=prefer_longer,
                try_convert_units=try_convert_units,
                rename_var=rename_var,
                **kwargs,
            )

        try:
            var_to_read = self._get_var_to_read(var_name)
            return self._load_var(
                var_name=var_to_read,
                ts_type=ts_type,
                start=start,
                stop=stop,
                experiment=experiment,
                vert_which=vert_which,
                flex_ts_type=flex_ts_type,
                prefer_longer=prefer_longer,
                try_convert_units=try_convert_units,
                rename_var=rename_var,
                **kwargs,
            )

        except VarNotAvailableError:
            if self.check_compute_var(var_name):
                return self.compute_var(
                    var_name=var_name,
                    start=start,
                    stop=stop,
                    ts_type=ts_type,
                    experiment=experiment,
                    vert_which=vert_which,
                    flex_ts_type=flex_ts_type,
                    prefer_longer=prefer_longer,
                    try_convert_units=try_convert_units,
                    rename_var=rename_var,
                    **kwargs,
                )
        # this input variable was explicitly set to be computed, in which
        # case reading of that variable is ignored even if a file exists for
        # that
        raise VarNotAvailableError(
            f"Error: variable {var_name} not available in files and can also not be computed."
        )

    def read(
        self,
        vars_to_retrieve=None,
        start=None,
        stop=None,
        ts_type=None,
        experiment=None,
        vert_which=None,
        flex_ts_type=True,
        prefer_longer=False,
        require_all_vars_avail=False,
        **kwargs,
    ):
        """Read all variables that could be found

        Reads all variables that are available (i.e. in :attr:`vars_filename`)

        Parameters
        ----------
        vars_to_retrieve : list or str, optional
            variables that are supposed to be read. If None, all variables
            that are available are read.
        start : Timestamp or str, optional
            start time of data import
        stop : Timestamp or str, optional
            stop time of data import
        ts_type : str, optional
            string specifying temporal resolution (choose from
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more
            than one experiment)
        vert_which : str or dict, optional
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel) or dictionary containing var_name as key and vertical
            coded string as value, accordingly
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that
            match the query.
        require_all_vars_avail : bool
            if True, it is strictly required that all input variables are
            available.
        **kwargs
            optional and support for deprecated input args

        Returns
        -------
        tuple
            loaded data objects (type :class:`GriddedData`)

        Raises
        ------
        IOError
            if input variable names is not list or string
        VarNotAvailableError
            1. if ``require_all_vars_avail=True`` and one or more of the
            desired variables is not available in this class
            2. if ``require_all_vars_avail=True`` and if none of the input
            variables is available in this object
        """
        if vars_to_retrieve is None and "var_names" in kwargs:
            warnings.warn(
                "Input arg var_names is deprecated. Please use vars_to_retrieve instead",
                DeprecationWarning,
                stacklevel=2,
            )
            vars_to_retrieve = kwargs["var_names"]
        if vars_to_retrieve is None:
            vars_to_retrieve = self.vars_filename
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        elif not isinstance(vars_to_retrieve, list):
            raise OSError(
                f"Invalid input for vars_to_retrieve {vars_to_retrieve}. "
                f"Need string or list of strings specifying var_names to load. "
                f"You may also leave it empty (None) in which case all "
                f"available variables are loaded"
            )
        if require_all_vars_avail:
            if not all([var in self.vars_provided for var in vars_to_retrieve]):
                raise VarNotAvailableError(
                    f"One or more of the specified vars ({vars_to_retrieve}) "
                    f"is not available in {self.data_id} database. "
                    f"Available vars: {self.vars_provided}"
                )
        var_names = list(np.intersect1d(self.vars_provided, vars_to_retrieve))
        if len(var_names) == 0:
            raise VarNotAvailableError(
                f"None of the desired variables is available in {self.data_id}"
            )
        data = []
        for var in var_names:
            try:
                data.append(
                    self.read_var(
                        var_name=var,
                        start=start,
                        stop=stop,
                        ts_type=ts_type,
                        experiment=experiment,
                        vert_which=vert_which,
                        flex_ts_type=flex_ts_type,
                        prefer_longer=prefer_longer,
                    )
                )
            except (VarNotAvailableError, DataCoverageError) as e:
                logger.warning(repr(e))
        return tuple(data)

    def _load_files(self, files, var_name, perform_fmt_checks=None):
        """Load list of files containing variable to read into Cube instances

        Parameters
        ----------
        files : list
            list of netcdf file
        var_name : str
            name of variable to read
        perform_fmt_checks : bool
            if True, the loaded data is checked for consistency with
            AeroCom default requirements.

        Returns
        -------
        CubeList
            list of loaded Cube instances
        list
            list containing corresponding filenames of loaded cubes
        """
        cubes, loaded_files = load_cubes_custom(
            files=files,
            var_name=var_name,
            file_convention=self.file_convention,
            perform_fmt_checks=perform_fmt_checks,
        )

        if len(loaded_files) == 0:
            raise OSError(f"None of the input files could be loaded in {self.data_id}")

        # self.loaded_cubes[var_name] = cubes
        return (cubes, loaded_files)

    def _get_meta_df(self, subset):
        """Extract relevant meta information from file_info dataframe

        Parameters
        ----------
        subset : pd.DataFrame
            subset of, or :attr:`file_info`

        Raises
        ------
        DataQueryError
            if more than one time frequency is available in the dataframe

        Returns
        -------

        """
        meta = {}
        # check additional metadata
        ts_types = subset.ts_type.unique()

        # sanity check
        if len(ts_types) > 1:
            raise DataQueryError("Fatal: subset contains more than one ts_type")
        meta["ts_type"] = ts_types[0]

        mets = subset.meteo.unique()
        meta["meteo"] = mets[0] if len(mets) == 1 else list(mets)

        exps = subset.experiment.unique()
        meta["experiment"] = exps[0] if len(exps) == 1 else list(exps)

        perts = subset.perturbation.unique()
        meta["perturbation"] = perts[0] if len(perts) == 1 else list(perts)

        vertcodes = subset.vert_code.unique()
        meta["vert_code"] = vertcodes[0] if len(vertcodes) == 1 else list(vertcodes)

        return meta

    def _load_var(
        self,
        var_name,
        ts_type,
        start,
        stop,
        experiment,
        vert_which,
        flex_ts_type,
        prefer_longer,
        try_convert_units,
        rename_var,
        **kwargs,
    ):
        """Find files corresponding to input specs and load into GriddedData

        Note
        ----
        See :func:`read_var` for I/O info.
        """
        if self.ignore_vert_code:
            vert_which = None
        subset = self.filter_query(
            var_name,
            ts_type,
            start,
            stop,
            experiment,
            vert_which,
            is_at_stations=False,
            flex_ts_type=flex_ts_type,
            prefer_longer=prefer_longer,
        )
        if len(subset) == 0:
            raise DataQueryError("Could not find file match for query")

        match_files = self._generate_file_paths(subset)
        (cube_list, from_files) = self._load_files(match_files, var_name, **kwargs)
        is_concat = False
        if len(cube_list) > 1:
            try:
                cube = self.concatenate_cubes(cube_list)
                is_concat = True
            except iris.exceptions.ConcatenateError as e:
                raise NotImplementedError(
                    f"Failed to concatenate cubes: {cube_list}\nError: {repr(e)}"
                )
        else:
            cube = cube_list[0]

        meta = self._get_meta_df(subset)
        data = GriddedData(
            input=cube,
            from_files=from_files,
            data_id=self.data_id,
            concatenated=is_concat,
            convert_unit_on_init=try_convert_units,
            **meta,
        )
        # crop cube in time (if applicable)
        if isinstance(start, int) and start != 9999:
            try:
                data = self._check_crop_time(data, start, stop)
            except Exception:
                logger.exception(
                    f"Failed to crop time dimension in {data} (start: {start}, stop: {stop})"
                )
        if rename_var is not None:
            data.var_name = rename_var
        data.reader = self
        return data

    def _check_crop_time(self, data, start, stop):
        crop_time = False
        crop_time_range = [self.start, self.stop]
        if start is not None:
            crop_time = True
            crop_time_range[0] = to_pandas_timestamp(start)
        if stop is not None:
            crop_time = True
            crop_time_range[1] = to_pandas_timestamp(stop)
        if crop_time:
            logger.info("Applying temporal cropping of result cube")
            data = data.crop(time_range=crop_time_range)
        return data

    def _check_ts_type(self, ts_type):
        """Check and, if applicable, update ts_type

        Returns
        -------
        str
            valid ts_type

        Raises
        ------
        ValueError

        """
        if ts_type is None:
            if len(self.ts_types) == 0:
                raise AttributeError(
                    "Apparently no files with a valid ts_type "
                    "entry in their filename could be found"
                )

            ts_type = self.ts_types[0]
        if not TsType.valid(ts_type):
            raise ValueError(f"Invalid input for ts_type: {ts_type}")
        return ts_type

    def __getitem__(self, var_name):
        """Try access import result for one of the models

        Parameters
        ----------
        var_name : str
            string specifying variable that is supposed to be extracted

        Returns
        -------
        GriddedData
            the corresponding read class for this model

        Raises
        -------
        ValueError
            if results for ``var_name`` are not available
        """
        if var_name not in self.data:
            return self.read_var(var_name)
        # return self.data[var_name]

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        head = f"Pyaerocom {type(self).__name__}"
        s = (
            f"\n{head}\n{len(head) * '-'}\n"
            f"Data ID: {self.data_id}\n"
            f"Data directory: {self.data_dir}\n"
            f"Available experiments: {self.experiments}\n"
            f"Available years: {self.years_avail}\n"
            f"Available frequencies {self.ts_types}\n"
            f"Available variables: {self.vars_provided}\n"
        )
        return s.rstrip()

    ### DEPRECATED STUFF
    @property
    def name(self):
        """Deprecated name of attribute data_id"""
        warnings.warn("Please use data_id", DeprecationWarning, stacklevel=2)
        return self.data_id


def is_3d(var_name: str) -> bool:
    return "3d" in var_name.lower()



from copy import deepcopy

import pytest

from pyaerocom.obs_io import (
    OBS_ALLOW_ALT_WAVELENGTHS,
    OBS_WAVELENGTH_TOL_NM,
    AuxInfoUngridded,
    ObsVarCombi,
)

AUX_EXAMPLE = dict(
    data_id="AERONET",
    vars_supported=["fmf550aer", "od550lt1aer"],
    aux_merge_how={"fmf550aer": "eval", "od550lt1aer": "combine"},
    aux_requires={
        "fmf550aer": {
            "AeronetSDAV3Lev2.daily": "od550lt1aer",
            "AeronetSunV3Lev2.daily": "od550aer",
        },
        "od550lt1aer": {
            "AeronetSDAV3Lev2.daily": "od550lt1aer",
            "AeronetSunV3Lev2.daily": "od550aer",
        },
    },
    aux_funs={
        "fmf550aer": "(AeronetSDAV3Lev2.daily;od550lt1aer/AeronetSunV3Lev2.daily;od550aer)*100"
    },
    aux_units={"fmf550aer": "%"},
)


def test_OBS_WAVELENGTH_TOL_NM():
    assert OBS_WAVELENGTH_TOL_NM == 10.0


def test_OBS_ALLOW_ALT_WAVELENGTHS():
    assert OBS_ALLOW_ALT_WAVELENGTHS is True


def test_ObsVarCombi():
    assert str(ObsVarCombi("bla", "blub")) == "bla;blub"


def test_AuxInfoUngridded_MAX_VARS_PER_METHOD():
    assert AuxInfoUngridded.MAX_VARS_PER_METHOD == 2


EX_WRONG1: dict = deepcopy(AUX_EXAMPLE)
EX_WRONG1.update(aux_funs=None)

EX_WRONG2 = deepcopy(AUX_EXAMPLE)
EX_WRONG2.update(vars_supported={"a": 1})

EX_WRONG3: dict = deepcopy(AUX_EXAMPLE)
EX_WRONG3["vars_supported"].append("blablub")

EX_WRONG4 = deepcopy(EX_WRONG3)
EX_WRONG4["aux_merge_how"].update(blablub="eval")

EX_WRONG5 = deepcopy(EX_WRONG4)
EX_WRONG5["aux_requires"].update(blablub=42)

EX_WRONG6 = deepcopy(EX_WRONG5)
EX_WRONG6["aux_requires"].update(blablub={"abc": "42"})

EX_WRONG7 = deepcopy(EX_WRONG6)
EX_WRONG7["aux_requires"]["blablub"].update({"def": "43"})


@pytest.mark.parametrize(
    "kwargs,error",
    [
        (EX_WRONG1, "Specification of computation function is missing for var fmf550aer"),
        (EX_WRONG2, "Variable a is not defined in attr aux_requires..."),
        (EX_WRONG3, "Variable blablub is not defined in attr aux_requires..."),
        (EX_WRONG4, "Variable blablub is not defined in attr aux_requires..."),
        (EX_WRONG5, "Specification of computation function is missing for var blablub"),
        (EX_WRONG6, "Specification of computation function is missing for var blablub"),
        (EX_WRONG7, "Specification of computation function is missing for var blablub"),
    ],
)
def test_AuxInfoUngridded_error(kwargs, error: str):
    with pytest.raises(ValueError) as e:
        AuxInfoUngridded(**kwargs)
    assert str(e.value) == error


EX_NOTWRONG1 = deepcopy(EX_WRONG7)
EX_NOTWRONG1["aux_funs"]["blablub"] = "abc;42+def;43"

EX_NOTWRONG2 = deepcopy(EX_NOTWRONG1)
EX_NOTWRONG2["aux_units"]["blablub"] = "1"


@pytest.mark.parametrize(
    "kwargs",
    [
        AUX_EXAMPLE,
        EX_NOTWRONG1,
        EX_NOTWRONG2,
    ],
)
def test_AuxInfoUngridded_to_dict(kwargs):
    info = AuxInfoUngridded(**kwargs)
    assert info.to_dict() == kwargs



from __future__ import annotations

import logging

import numpy as np
import simplejson

logger = logging.getLogger(__name__)


FLOAT_DECIMALS = 5


def set_float_serialization_precision(precision: int) -> None:
    """update `FLOAT_DECIMALS`"""
    global FLOAT_DECIMALS
    FLOAT_DECIMALS = precision


def round_floats(in_data: float | dict | list | tuple) -> float | dict | list | tuple:
    """
    round all floats in `in_data` to `FLOAT_DECIMALS` precision.
    For nested structures, this method is called recursively to go through
    all levels

    Parameters
    ----------
    in_data : float, dict, tuple, list
        data structure whose numbers should be limited in precision

    Returns
    -------
    in_data
        all the floats in in_data with limited precision
        tuples in the structure have been converted to lists to make them mutable

    """

    if isinstance(in_data, float | np.float32 | np.float16 | np.float128 | np.float64):
        # np.float64, is an alias for the Python float, but is mentioned here for completeness
        # note that round and np.round yield different results with the Python round being mathematically correct
        # details are here:
        # https://numpy.org/doc/stable/reference/generated/numpy.around.html#numpy.around
        # use numpy around for now
        return np.around(in_data, FLOAT_DECIMALS)
    elif isinstance(in_data, list | tuple):
        return [round_floats(v) for v in in_data]
    elif isinstance(in_data, dict):
        return {k: round_floats(v) for k, v in in_data.items()}
    return in_data


def read_json(file_path):
    """Read json file

    Parameters
    ----------
    file_path : str
        json file path

    Returns
    -------
    dict
        content as dictionary
    """
    with open(file_path) as f:
        data = simplejson.load(f, allow_nan=True)
    return data


def write_json(data_dict, file_path, **kwargs):
    """Save json file

    Parameters
    ----------
    data_dict : dict
        dictionary that can be written to json file
    file_path : str
        output file path
    **kwargs
        additional keyword args, e.g.
        indent=int
        (ignore_nan, forced to True)
        round_floats=True/False round floating numbers, default True
    """
    kwargs.update(ignore_nan=True)
    if kwargs.pop("round_floats", True):
        data_dict = round_floats(in_data=data_dict)
    with open(file_path, "w") as f:
        simplejson.dump(data_dict, f, allow_nan=True, **kwargs)



import logging

import numpy as np
from geonum.atmosphere import T0_STD, p0

from pyaerocom.io.aux_read_cubes import (
    CUBE_MATHS,
    _check_input_iscube,
    _check_same_units,
    add_cubes,
)
from pyaerocom.units.molecular_mass import get_molmass

from pyaerocom.units import Unit


logger = logging.getLogger(__name__)


single_component_mass = {"n": 14.0067, "c": 12.011, "s": 32.065}


def vmr_to_conc(data, vmr_unit, var_name, to_unit, component_unit=None):
    """
    Convert volume mixing ratio (vmr) to mass concentration

    Parameters
    ----------
    data : cube or GriddedData
        array containing vmr values

    vmr_unit : str
        unit of input data
    var_name: str
        name of variable, used to fund molar mass
    to_unit : str, optional
        Unit to which output data is converted. If None, output unit is
        kg m-3. The default is None.
    component_unit : str, optional
        If none, the to_unit unit is returned. If, e.g. n or N, then ug m-3 -> ug N m-3 is returned

    Returns
    -------
    cube
        input data converted to mass concentration

    """

    data = _check_input_iscube(data)[0]

    p_pascal = p0  # 1013 hPa (US standard atm)
    T_kelvin = T0_STD  # 15 deg celsius (US standard atm)

    mmol_air = get_molmass("air_dry")
    mmol_var = get_molmass(var_name)

    if component_unit is not None and to_unit is not None:
        component_mass = single_component_mass[component_unit.lower()]
        component_unit_fac = component_mass / mmol_var
    else:
        component_unit_fac = 1
    Rspecific = 287.058  # J kg-1 K-1

    conversion_fac = 1 / Unit("mol mol-1").convert(1, vmr_unit)

    airdensity = p_pascal / (Rspecific * T_kelvin)  # kg m-3
    mulfac = mmol_var / mmol_air * airdensity  # kg m-3

    mult_fun = CUBE_MATHS["multiply"]
    conc = mult_fun(data, mulfac)  # kg m-3
    if to_unit is not None:
        conversion_fac *= Unit("kg m-3").convert(1, to_unit) * component_unit_fac
    if not np.isclose(conversion_fac, 1, rtol=1e-7):
        conc = mult_fun(conc, conversion_fac)

    if to_unit is not None:
        unit = to_unit

        if component_unit is not None:
            unit_list = unit.split(" ")
            unit = unit_list[0] + f" {component_unit.upper()} " + unit_list[1]

        conc.units = unit
    else:
        conc.units = "kg m-3"

    return conc


def calc_concNhno3_from_vmr(data):
    return vmr_to_conc(
        data, vmr_unit="nmol mol-1", var_name="hno3", to_unit="ug m-3", component_unit="N"
    )


def calc_concNnh3_from_vmr(data):
    return vmr_to_conc(
        data, vmr_unit="nmol mol-1", var_name="nh3", to_unit="ug m-3", component_unit="N"
    )


def convert_to_ugN(data, var_name):
    mult_fun = CUBE_MATHS["multiply"]
    data = _check_input_iscube(data)[0]
    mmol_var = get_molmass(var_name)

    component_mass = single_component_mass["n"]
    component_unit_fac = component_mass / mmol_var
    unit = data.units
    unit_conversion = Unit(str(unit)).convert(1, "ug m-3")
    if not np.isclose(unit_conversion, 1, rtol=1e-7):
        data = mult_fun(data, unit_conversion)

    data = mult_fun(data, component_unit_fac)
    data.units = "ug N m-3"

    return data


def calc_concNnh4(concnh4):
    return convert_to_ugN(concnh4, "nh4")


def calc_concno3pm25(concno3f, concno3c, fine_from_coarse_fraction: float = 0.134):
    # mult_fun = CUBE_MATHS["multiply"]
    # concno3pm25 = add_cubes(concno3f, mult_fun(concno3c, fine_from_coarse_fraction))

    return concno3f


def calc_concno3pm10(concno3f, concno3c):
    concno3pm10 = add_cubes(concno3f, concno3c)

    return concno3pm10


def calc_concNno3pm25(concno3f, concno3c, fine_from_coarse_fraction: float = 0.134):
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + M_O * 3)
    mult_fun = CUBE_MATHS["multiply"]
    concno3f, concno3c = _check_input_iscube(concno3f, concno3c)
    concno3f, concno3c = _check_same_units(concno3f, concno3c)
    concno3pm25 = add_cubes(concno3f, mult_fun(concno3c, fine_from_coarse_fraction))
    concno3pm25.units = "ug N m-3"
    return mult_fun(concno3pm25, fac)


def calc_concNno3pm10(concno3f, concno3c):
    M_N = 14.006
    M_O = 15.999

    fac = M_N / (M_N + M_O * 3)
    mult_fun = CUBE_MATHS["multiply"]
    concno3f, concno3c = _check_input_iscube(concno3f, concno3c)
    concno3f, concno3c = _check_same_units(concno3f, concno3c)
    concno3pm10 = add_cubes(concno3f, concno3c)
    concno3pm10.units = "ug N m-3"
    return mult_fun(concno3pm10, fac)


def calc_sspm25(concssfine, concsscoarse):
    mult_fun = CUBE_MATHS["multiply"]
    concssfine, concsscoarse = _check_input_iscube(concssfine, concsscoarse)
    concssfine, concsscoarse = _check_same_units(concssfine, concsscoarse)

    return add_cubes(concssfine, mult_fun(concsscoarse, 0.16))


def calc_concNtno3(concno3f, concno3c, vmrhno3):
    concno3f, concno3c, vmrhno3 = _check_input_iscube(concno3f, concno3c, vmrhno3)
    concno3f, concno3c = _check_same_units(concno3f, concno3c)
    concNhno3 = calc_concNhno3_from_vmr(vmrhno3)
    concNno3pm10 = calc_concNno3pm10(concno3f, concno3c)

    return add_cubes(concNhno3, concNno3pm10)


def calc_concNtnh(concnh4, vmrnh3):
    concNnh3 = calc_concNnh3_from_vmr(vmrnh3)
    concNnh4 = calc_concNnh4(concnh4)
    concNnh3, concNnh4 = _check_same_units(concNnh3, concNnh4)

    return add_cubes(concNnh3, concNnh4)



from __future__ import annotations

import logging
import re
from collections import defaultdict
from collections.abc import Iterable
from functools import cached_property, lru_cache
from pathlib import Path

import xarray as xr

from pyaerocom import const
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

from .aux_vars import vmrno2_from_ds, vmro3_from_ds, vmro3max_from_ds

logger = logging.getLogger(__name__)

COLUMNS = (
    "country",
    "stationname",
    "stationcode",
    "latitude",
    "longitude",
    "altitude",
    "timestamp",
    "pollutant",
    "unit",
    "frequency",
    "value",
)
METADATA_INDEX_START = 11
ALLOWED_FREQS = {
    "hourly",
    "daily",
    "monthly",
    "yearly",
}


class ReadCNEMC(ReadUngriddedBase):
    """Class for reading CNEMC observations (formerly MEP and before that MarcoPolo)

    Note
    ----
    Currently only single variable reading into an :class:`UngriddedData`
    object is supported.
    """

    #: Mask for identifying datafiles
    _FILEMASK = "mep-rd-*.nc"

    #: Version log of this class (for caching)
    __version__ = "0.02"

    #: Name of the dataset (OBS_ID)
    DATA_ID = const.CNEMC_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: There is no global ts_type but it is specified in the data files...
    TS_TYPE = "variable"

    #: sampling frequencies found in data files
    TS_TYPES_FILE = {"hour": "hourly", "day": "daily"}

    #: field name of the start time of the measurement (in lower case)
    START_TIME_NAME = "datetime_start"

    #: filed name of the end time of the measurement (in lower case)
    END_TIME_NAME = "datetime_stop"

    #: there's no general instrument name in the data
    INSTRUMENT_NAME = "unknown"

    DATA_PRODUCT = ""

    #: functions used to convert variables that are computed
    AUX_FUNS = {
        "vmro3": vmro3_from_ds,
        "vmro3max": vmro3max_from_ds,
        "vmrno2": vmrno2_from_ds,
    }

    VAR_MAPPING = {
        "concco": "CO_density",
        "concno2": "NO2_density",
        "conco3": "O3_density",
        "concpm10": "PM10_density",
        "concpm25": "PM2p5_density",
        "concso2": "SO2_density",
    }

    STATION_REGEX = re.compile("mep-rd-(.*A)-.*.nc")

    DEFAULT_VARS = list(VAR_MAPPING)

    DATASET_NAME = DATA_ID

    PROVIDES_VARIABLES = list(VAR_MAPPING) + list(AUX_FUNS)

    def __init__(self, data_id: str | None = None, data_dir: str | None = None):
        if data_dir is None:
            data_dir = const.OBSLOCS_UNGRIDDED[const.CNEMC_NAME]

        super().__init__(data_id=data_id, data_dir=data_dir)
        self.files = sorted(map(str, self.FOUND_FILES))

    @cached_property
    def FOUND_FILES(self) -> tuple[Path, ...]:
        paths = sorted(Path(self.data_dir).rglob(self._FILEMASK))
        logger.debug(f"found {len(paths)} files")
        return tuple(paths)

    @lru_cache
    def stations(self, files: tuple[str | Path, ...] | None = None) -> dict[str, list[Path]]:
        if not files:
            files = self.FOUND_FILES

        stations = defaultdict(list)
        for path in files:
            if not isinstance(path, Path):
                path = Path(path)
            if (name := self._station_name(path)) is None:
                logger.debug(f"Skipping {path.name}")
                continue
            stations[name].append(path)

        logger.debug(f"found {len(stations)} stations")
        return stations

    @classmethod
    def _station_name(cls, path: Path) -> str | None:
        match = cls.STATION_REGEX.search(path.name)
        return match.group(1) if match else None

    def read_file(
        self, filename: str | Path, vars_to_retrieve: Iterable[str] | None = None
    ) -> UngriddedData:
        """Reads data for a single year for one component"""
        if not isinstance(filename, Path):
            filename = Path(filename)
        if not filename.is_file():
            raise ValueError(f"missing {filename}")
        return self.read(vars_to_retrieve, (filename,))

    def read(
        self,
        vars_to_retrieve: Iterable[str] | None = None,
        files: Iterable[str | Path] | None = None,
        first_file: int | None = None,
        last_file: int | None = None,
        metadatafile=None,
    ) -> UngriddedData:
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS

        if unsupported := set(vars_to_retrieve) - set(self.PROVIDES_VARIABLES):
            raise ValueError(f"Unsupported variables: {', '.join(sorted(unsupported))}")

        if files is not None and not isinstance(files, tuple):
            files = tuple(files)

        if files is not None and first_file is not None:
            files = files[first_file:]

        if files is not None and last_file is not None:
            files = files[:last_file]

        stations: list[StationData] = []

        for station_name, paths in self.stations(files).items():
            logger.debug(f"Reading station {station_name}")
            ds = self._read_dataset(paths)[vars_to_retrieve]
            stations.append(self.to_stationdata(ds, station_name))

        return UngriddedData.from_station_data(stations)

    def _read_dataset(self, paths: list[Path]) -> xr.Dataset:
        ds = xr.open_mfdataset(
            sorted(paths),
            concat_dim="time",
            combine="nested",
            parallel=True,
            decode_cf=True,
            decode_timedelta=True,
        )
        ds = ds.rename({v: k for k, v in self.VAR_MAPPING.items()})
        ds = ds.assign(
            time=self._dataset_time(ds),
            **{name: func(ds) for name, func in self.AUX_FUNS.items()},
        )
        return ds.set_coords(("latitude", "longitude", "altitude"))

    @classmethod
    def _dataset_time(cls, ds: xr.Dataset) -> xr.DataArray:
        # can not add ds["datetime_start"] and ds["datetime_start"], as both are of type datetime[ns]
        time = ds[cls.START_TIME_NAME] + (ds[cls.END_TIME_NAME] - ds[cls.START_TIME_NAME]) / 2
        return xr.Variable(
            "time",
            time,
            dict(
                long_name="time at middle of the period",
                units=ds[cls.START_TIME_NAME].encoding["units"],
            ),
            ds[cls.START_TIME_NAME].encoding,
        )

    @classmethod
    def to_stationdata(cls, ds: xr.Dataset, station_name: str) -> StationData:
        station = StationData()
        station.data_id = cls.DATA_ID
        station.dataset_name = cls.DATASET_NAME
        station.station_id = station_name
        station.station_name = station_name
        station.latitude = float(ds["latitude"][0])
        station.longitude = float(ds["longitude"][0])
        station.altitude = float(ds["altitude"][0])

        station.station_coords = {
            "latitude": station.latitude,
            "longitude": station.longitude,
            "altitude": station.altitude,
        }

        station.ts_type = "hourly"
        station["dtime"] = ds["time"].values

        for var in ds.data_vars:
            station[var] = ds[var].to_series()
            station["var_info"][var] = {"units": ds[var].units}

        return station



import numpy as np
from pyaerocom.dynamic_rec_array import DynamicRecArray


my_dtype = [("int", "i4"), ("int2", "i2")]


def test_DynamicRecArray__init__():
    ary = DynamicRecArray(my_dtype)
    assert ary._capacity == 10  # default capacity
    assert len(ary._array) == ary._capacity
    assert len(ary) == 0
    # internal structure on create
    assert isinstance(ary._array, np.ndarray)
    assert ary._array["int"].shape[0] == ary._capacity
    assert ary._array["int2"].shape[0] == ary._capacity
    # external structure should be empty
    assert isinstance(ary.data, np.ndarray)
    assert ary.data["int"].shape[0] == 0
    assert ary.data["int2"].shape[0] == 0

    ary = DynamicRecArray(my_dtype, 100)
    assert ary._capacity == 100
    assert len(ary._array) == ary._capacity
    assert len(ary) == 0

    # keys
    assert set(ary.keys()) == set([x[0] for x in my_dtype])


def test_data():
    ary = DynamicRecArray(my_dtype, 10)
    assert len(ary) == 0
    ary2 = DynamicRecArray(my_dtype, 100)
    rec_array = ary2._array
    ary.data = rec_array
    assert len(ary) == 100


def test_append():
    ary = DynamicRecArray(my_dtype, 10)
    assert len(ary) == 0
    # append single row
    ary.append([(0, 1)])
    assert len(ary) == 1
    assert ary.data[0]["int"] == 0
    assert ary.data[0]["int2"] == 1

    ary2 = DynamicRecArray(my_dtype, 100)
    rec_array = ary2._array
    rec_array["int"] = np.arange(100)
    rec_array["int2"] = 100 * rec_array["int"]
    ary2.data = rec_array
    assert len(ary2) == 100
    assert ary2.data[99]["int"] == 99
    assert ary2.data[99]["int2"] == 99 * 100

    ary.append(rec_array)
    assert len(ary) == 101
    assert ary.data[0]["int"] == 0
    assert ary.data[0]["int2"] == 1
    assert ary.data[100]["int"] == 99
    assert ary.data[100]["int2"] == 99 * 100


def test_append_array():
    ary = DynamicRecArray(my_dtype, 10)
    ary.append([(0, 1)])
    assert len(ary) == 1
    assert ary.data[0]["int"] == 0
    assert ary.data[0]["int2"] == 1

    a1 = np.zeros(100) + 1
    a2 = np.zeros(100) + 2
    ary.append_array(int=a1, int2=a2)
    assert len(ary) == 101
    assert ary.data[100]["int"] == 1
    assert ary.data[100]["int2"] == 2



from .units import Unit
from typing import TypeVar
from pyaerocom.units.exceptions import UnitConversionError
from pyaerocom.units.units import UnitConversionCallbackHandler

from .typing import SupportsMul

__all__ = ["get_unit_conversion_fac", "convert_unit"]

T = TypeVar("T", bound=SupportsMul)
#: default frequency for rates variables (e.g. deposition, precip)
RATES_FREQ_DEFAULT = "d"


def get_unit_conversion_fac(
    from_unit: str, to_unit: str, var_name: str | None = None, ts_type: str | None = None
) -> float:
    """Gets a unit conversion factor for converting from one unit to another.

    :param from_unit: From unit.
    :param to_unit: To unit.
    :param var_name: aerocom var name, defaults to None
    :param ts_type: ts_type, defaults to None

    :return: Conversion factor.

    :raises: UnitConversionError
        if unable to convert between units.
    """
    return convert_unit(
        1, from_unit=from_unit, to_unit=to_unit, var_name=var_name, ts_type=ts_type
    )


def convert_unit(
    data: T,
    /,
    from_unit: str,
    to_unit: str,
    var_name: str | None = None,
    ts_type: str | None = None,
    *,
    inplace: bool = False,
    callback: UnitConversionCallbackHandler | None = None,
) -> T:
    """Convert unit of data

    Parameters
    ----------
    data : np.ndarray or similar
        input data
    from_unit : cf_units.Unit or str
        current unit of input data
    to_unit : cf_units.Unit or str
        new unit of input data
    var_name : str, optional
        name of variable. If provided, and standard conversion with
        :mod:`cf_units` fails, then custom unit conversion is attempted.
    ts_type : str, optional
        frequency of data. May be needed for conversion of rate variables
        such as precip, deposition, etc, that may be defined implicitly
        without proper frequency specification in the unit string.

    Returns
    -------
    data
        data in new unit
    """
    try:
        data = Unit(from_unit, aerocom_var=var_name, ts_type=ts_type).convert(
            data,
            other=Unit(to_unit, aerocom_var=var_name, ts_type=ts_type),
            inplace=inplace,
            callback=callback,
            aerocom_var=var_name,
        )
    except ValueError as e:
        raise UnitConversionError(
            f"failed to convert unit from {str(from_unit)} to {to_unit}"
        ) from e

    return data



from datetime import date

import numpy as np
import pytest

from pyaerocom.trends_helpers import (
    _end_season,
    _find_area,
    _get_season_from_months,
    _init_trends_result_dict,
    _mid_season,
    _start_season,
    _start_stop_period,
    _years_from_periodstr,
)


def test__init_trends_result_dict():
    result = _init_trends_result_dict(2022)
    assert isinstance(result, dict)
    assert len(result.keys()) == 14


def test__get_season_from_months_error():
    with pytest.raises(ValueError):
        _get_season_from_months(months="cat")


@pytest.mark.parametrize(
    "season, year",
    [
        ("spring", 2022),
        ("summer", 2022),
        ("autumn", 2022),
        ("winter", 2022),
        ("all", 2022),
    ],
)
def test__mid_season(season, year):
    res = _mid_season(season, year)
    assert isinstance(res, np.datetime64)
    assert "15" in str(res)


def test__mid_season_error():
    with pytest.raises(ValueError):
        _mid_season(seas="cat", yr=2022)


@pytest.mark.parametrize(
    "season, year",
    [
        ("spring", 2022),
        ("summer", 2022),
        ("autumn", 2022),
        ("winter", 2022),
        ("all", 2022),
    ],
)
def test__start_season(season, year):
    res = _start_season(season, year)
    assert isinstance(res, str)
    assert "01" in res


def test__start_season_error():
    with pytest.raises(ValueError):
        _start_season("cat", 2022)


@pytest.mark.parametrize(
    "season, year",
    [
        ("spring", 2022),
        ("summer", 2022),
        ("autumn", 2022),
        ("winter", 2022),
        ("all", 2022),
    ],
)
def test__end_season(season, year):
    res = _end_season(season, year)
    assert isinstance(res, str)
    assert "01" in res


def test__end_season_error():
    with pytest.raises(ValueError):
        _end_season(seas="cat", yr=2022)


@pytest.mark.parametrize(
    "lat, lon, area",
    [
        (30.033333, 31.233334, "NAFRICA"),
        (48.8566, 2.35, "EUROPE"),
        (22.572645, 88.363892, "INDIA"),
    ],
)
def test__find_area(lat, lon, area):
    res = _find_area(lat, lon)
    assert area == res


def test__years_from_periodstr():
    period = "1990-2010"
    res = _years_from_periodstr(period)
    assert len(res) == 2
    assert all(isinstance(item, int) for item in res)


def test__start_stop_period():
    period = "1990-2010"
    res = _start_stop_period(period)
    assert len(res) == 2
    assert all(isinstance(item, date) for item in res)



import logging
import os

import numpy as np

from pyaerocom.aeroval import EvalSetup, ExperimentProcessor
from pyaerocom.griddeddata import GriddedData
from pyaerocom.helpers import make_dummy_cube_latlon, numpy_to_cube
from pyaerocom.variable_helpers import get_variable

logger = logging.getLogger(__name__)


def make_config_template(proj_id: str, exp_id: str) -> EvalSetup:
    """
    Make a template for an AeroVal evaluation setup

    Parameters
    ----------
    proj_id : str
        ID of project.
    exp_id : str
        ID of experiment

    Returns
    -------
    EvalSetup
        template evaluation setup (all defaults are set) that can be used to
        add model and obs entries, and modified to meet the purposes.
    """
    return EvalSetup(proj_id, exp_id)


def compute_model_average_and_diversity(
    cfg,
    var_name,
    model_names=None,
    ts_type=None,
    lat_res_deg=2,
    lon_res_deg=3,
    data_id=None,
    avg_how=None,
    extract_surface=True,
    ignore_models=None,
    comment=None,
    model_use_vars=None,
):
    """Compute median or mean model based on input models

    Note
    ----
    - BETA version that will likely undergo revisions.
    - Time selection currently not properly handled

    Parameters
    ----------
    cfg : AerocomEvaluation
        analysis instance
    var_name : str
        name of variable
    model_names : list, optional
        list of model names. If None, all entries in input engine are used.
    ts_type : str, optional
        output freq. Defaults to monthly.
    lat_res_deg : int, optional
        output latitude resolution, defaults to 2 degrees.
    lon_res_deg : int, optional
        output longitude resolution, defaults to 3 degrees.
    data_id : str, optional
        output data_id of ensemble model.
    avg_how : str, optional
        how to compute averages (choose from mean or median), defaults to
        "median".
    extract_surface : bool
        if True (and if data contains model levels), surface level is
        extracted
    ignore_models : list, optional
        list of models to be ignored
    comment : str, optional
        comment string added to metadata of output data objects.
    model_use_vars : dict, optional
        model variables to be used.


    Returns
    -------
    GriddedData
        ensemble model for input variable computed averaged using median or
        mean (input avg_how). Default is median.
    GriddedData
        corresponding diversity field, if avg_how is "mean", then computed
        using definition from
        Textor et al., 2006 (ACP) DOI: 10.5194/acp-6-1777-2006. If avg_how
        is "median" then interquartile range is used (Q3-Q1)/Q2
    GriddedData or None
        Q1 field (only output if avg_how is median)
    GriddedData or None
        Q3 field (only output if avg_how is median)
    GriddedData or None
        standard deviation field (only output if avg_how is mean)
    """
    if not isinstance(cfg, ExperimentProcessor):
        raise ValueError("invalid input, need ExperimentProcessor")
    cfg.cfg._check_time_config()
    if ts_type is None:
        ts_type = "monthly"
    if avg_how is None:
        avg_how = "median"
    if model_use_vars is None:
        model_use_vars = {}
    if ignore_models is None:
        ignore_models = []

    if avg_how == "mean":
        avg_fun = np.mean
    elif avg_how == "median":
        avg_fun = np.median
    else:
        raise ValueError(f"Invalid input for avg_how {avg_how}")

    if data_id is None:
        data_id = f"AEROCOM-{avg_how.upper()}"

    if model_names is None:
        model_names = list(cfg.cfg.model_cfg)

    if len(ignore_models) > 0:
        models = []
        for model in model_names:
            if model not in ignore_models:
                models.append(model)
    else:
        models = model_names

    if not len(models) > 1:
        raise ValueError("Need more than one model to compute average...")

    dummy = make_dummy_cube_latlon(lat_res_deg=lat_res_deg, lon_res_deg=lon_res_deg)

    loaded = []
    from_files = []
    from_models = []
    from_vars = []
    models_failed = []

    unit_out = get_variable(var_name).units

    for m in models:
        mname = m.model_name
        logger.info(f"Adding {mname} ({var_name})")

        mid = cfg.cfg.model_cfg.get_entry(mname).model_id
        if mid == data_id or mname == data_id:
            continue

        read_var = var_name
        col = cfg.get_colocator(model_name=mname)
        if mname in model_use_vars:
            muv = model_use_vars[mname]
            if var_name in muv:
                read_var = muv[var_name]

        try:
            data = col.get_model_data(read_var)
            if not data.units == unit_out:
                data.convert_unit(unit_out)

            elif not data.longitude.circular:
                if not data.check_lon_circular():
                    raise Exception(f"Longitude of {mname} is not circular...")
            data.reorder_dimensions_tseries()
            data = data.resample_time(ts_type)
            if data.ndim == 4:
                if extract_surface:
                    data = data.extract_surface_level()
                else:
                    raise NotImplementedError("Cannot process ModelLevel fields yet")
            data = data.regrid(dummy)
            logger.info("Success!")
        except Exception as e:
            models_failed.append(mid)
            logger.info(f"Failed! Reason: {e}")
            continue

        loaded.append(data.cube.data)
        from_files.extend(data.from_files)
        from_models.append(data.data_id)
        from_vars.append(data.var_name)

    if not len(loaded) > 1:
        raise ValueError("Can only compute average if more than one model is available")

    from_files = [os.path.basename(f) for f in from_files]

    dims = [data.time, dummy.coord("latitude"), dummy.coord("longitude")]

    # the merged data objects
    arr = np.asarray(loaded)

    # average (mean or median)
    avgarr = avg_fun(arr, axis=0)

    if avg_how == "mean":
        stdarr = np.std(arr, axis=0)
        divarr = np.std(arr / avgarr, axis=0) * 100
    else:
        q1arr = np.quantile(arr, 0.25, axis=0)
        q3arr = np.quantile(arr, 0.75, axis=0)
        divarr = (q3arr - q1arr) / avgarr * 100

    if comment is None:
        comment = f"Ensemble {avg_how} for variable {var_name}. "

    # median or mean
    avg_out = GriddedData(
        numpy_to_cube(
            avgarr,
            dims=dims,
            var_name=var_name,
            units=data.units,
            ts_type=ts_type,
            data_id=data_id,
            from_files=from_files,
            from_models=from_models,
            from_vars=from_vars,
            models_failed=models_failed,
            comment=comment,
        )
    )

    commentdiv = comment + " Diversity field in units of % (IQR for median, std for mean)"

    # IQR or std based diversity
    div_out = GriddedData(
        numpy_to_cube(
            divarr,
            dims=dims,
            var_name=f"{var_name}div",
            units="%",
            ts_type=ts_type,
            data_id=data_id,
            from_files=from_files,
            from_models=from_models,
            from_vars=from_vars,
            models_failed=models_failed,
            comment=commentdiv,
        )
    )

    std_out, q1_out, q3_out = None, None, None
    if avg_how == "mean":
        commentstd = comment + " Standard deviation"
        std_out = GriddedData(
            numpy_to_cube(
                stdarr,
                dims=dims,
                var_name=f"{var_name}std",
                units=data.units,
                ts_type=ts_type,
                data_id=data_id,
                from_files=from_files,
                from_models=from_models,
                from_vars=from_vars,
                models_failed=models_failed,
                comment=commentstd,
            )
        )
    else:
        commentq1 = comment + " First quantile."
        q1_out = GriddedData(
            numpy_to_cube(
                q1arr,
                dims=dims,
                var_name=f"{var_name}q1",
                units=data.units,
                ts_type=ts_type,
                data_id=data_id,
                from_files=from_files,
                from_models=from_models,
                from_vars=from_vars,
                models_failed=models_failed,
                comment=commentq1,
            )
        )

        commentq3 = comment + " Third quantile."
        q3_out = GriddedData(
            numpy_to_cube(
                q3arr,
                dims=dims,
                var_name=f"{var_name}q3",
                units=data.units,
                ts_type=ts_type,
                data_id=data_id,
                from_files=from_files,
                from_models=from_models,
                from_vars=from_vars,
                models_failed=models_failed,
                comment=commentq3,
            )
        )

    return (avg_out, div_out, q1_out, q3_out, std_out)
